def getERC20ContractAddress(token_symbol):
    """
    Input an ERC20 token symbol, return the contract address
    """
    
    known_ERC20_token_symbols = [
        {
            "address": "0x4E84E9e5fb0A972628Cf4568c403167EF1D40431",
            "symbol": "$FFC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa024e8057eec474a9b2356833707dd0579e26ef3",
            "symbol": "$FYX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcdb7ecfd3403eef3882c65b761ef9b5054890a47",
            "symbol": "$HUR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0db8d8b76bc361bacbb72e2c491e06085a97ab31",
            "symbol": "$IQN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7dd7f56d697cc0f2b52bd55c057f378f1fe6ab4b",
            "symbol": "$TEAK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB6eD7644C69416d67B522e20bC294A9a9B405B31",
            "symbol": "0xBTC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xAf30D2a7E90d7DC361c8C4585e9BB7D2F6f15bc7",
            "symbol": "1ST",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfdbc1adc26f0f8f8606a5d63b7d3a3cd21c22b23",
            "symbol": "1WO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x0073e5E52E2B4fE218D75d994eE2B3c82f9C87EA",
            "symbol": "22x",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x9fC0583220eB44fAeE9e2dc1E63F39204DDD9090",
            "symbol": "2DC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xaEc98A708810414878c3BCDF46Aad31dEd4a4557",
            "symbol": "300",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x430241368c1D293fdA21DBa8Bb7aF32007c59109",
            "symbol": "3LT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xBa7DCBa2Ade319Bc772DB4df75A76BA00dFb31b0",
            "symbol": "A18",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb98d4c97425d9908e66e53a6fdf673acca0be986",
            "symbol": "ABT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0e8d6b471e332f140e7d9dbb99e5e3822f728da6",
            "symbol": "ABYSS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x13f1b7fdfbe1fc66676d56483e21b1ecb40b58e2",
            "symbol": "ACC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x06147110022b768ba8f99a8f385df11a151a9cc8",
            "symbol": "ACE",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x2baac9330cf9ac479d819195794d79ad0c7616e3",
            "symbol": "ADB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE69a353b3152Dd7b706ff7dD40fe1d18b7802d31",
            "symbol": "ADH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8810C63470d38639954c6B41AaC545848C46484a",
            "symbol": "ADI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x660e71483785f66133548b10f6926dc332b06e61",
            "symbol": "ADL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x422866a8F0b032c5cf1DfBDEf31A20F4509562b0",
            "symbol": "ADST",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xD0D6D6C5Fe4a677D343cC433536BB717bAe167dD",
            "symbol": "ADT",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x4470BB87d77b963A013DB939BE332f927f2b992e",
            "symbol": "ADX",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x5ca9a71b1d01849c0a95490cc00559717fcf0d1d",
            "symbol": "AE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8eB24319393716668D768dCEC29356ae9CfFe285",
            "symbol": "AGI",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x5121e348e897daef1eef23959ab290e5557cf274",
            "symbol": "AI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x37e8789bb9996cac9156cd5f5fd32599e6b91289",
            "symbol": "AID",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4CEdA7906a5Ed2179785Cd3A40A69ee8bc99C466",
            "symbol": "AION",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x27dce1ec4d3f72c3e457cc50354f1f975ddef488",
            "symbol": "AIR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1063ce524265d5a3A624f4914acd573dD89ce988",
            "symbol": "AIX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1ca43a170bad619322e6f54d46b57e504db663aa",
            "symbol": "AKC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x181a63746d3adcf356cbc73ace22832ffbb1ee5a",
            "symbol": "ALCO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x4289c043a12392f1027307fb58272d8ebd853912",
            "symbol": "ALI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xEA610B1153477720748DC13ED378003941d84fAB",
            "symbol": "ALIS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x638ac149ea8ef9a1286c41b977017aa7359e6cfa",
            "symbol": "ALTS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x49b127bc33ce7e1586ec28cec6a65b112596c822",
            "symbol": "ALX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4dc3643dbc642b72c158e7f3d2ff232df61cb6ce",
            "symbol": "AMB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x949bed886c739f1a3273629b3320db0c5024c719",
            "symbol": "AMIS",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0xca0e7269600d353f70b14ad118a49575455c0f2f",
            "symbol": "AMLT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x737f98ac8ca59f2c68ad658e3c3d8c8963e40a4c",
            "symbol": "AMN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x38c87aa89b2b8cd9b95b736e1fa7b612ea972169",
            "symbol": "AMO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x84936cF7630AA3e27Dd9AfF968b140d5AEE49F5a",
            "symbol": "AMTC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x960b236A07cf122663c4303350609A66A7B288C0",
            "symbol": "ANT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9ab165d795019b6d8b3e971dda91071421305e5a",
            "symbol": "AOA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4c0fbe1bb46612915e7967d2c3213cd4d87257ad",
            "symbol": "APIS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1a7a8bd9106f2b8d977e08582dc7d24c723ab0db",
            "symbol": "APPC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x23ae3c5b39b12f0693e05435eeaa1e51d8c61530",
            "symbol": "APT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xafbec4d65bc7b116d85107fd05d912491029bf46",
            "symbol": "ARB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xAc709FcB44a43c35F0DA4e3163b117A17F3770f5",
            "symbol": "ARC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1245ef80f4d9e02ed9425375e8f649b9221b31d8",
            "symbol": "ARCT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x75aa7b0d02532f3833b66c7f0ad35376d373ddf8",
            "symbol": "ARD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xBA5F11b16B155792Cf3B2E6880E8706859A8AEB6",
            "symbol": "ARN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xfec0cF7fE078a500abf15F1284958F22049c2C7e",
            "symbol": "ART",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7705FaA34B16EB6d77Dfc7812be2367ba6B0248e",
            "symbol": "ARX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xb0D926c1BC3d78064F3e1075D5bD9A24F35Ae6C5",
            "symbol": "ARXT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa5f8fc0921880cb7342368bd128eb8050442b1a1",
            "symbol": "ARY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x27054b13b1B798B345b591a4d22e6562d47eA75a",
            "symbol": "AST",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x7b22938ca841aa392c93dbb7f4c42178e3d65e88",
            "symbol": "ASTRO",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x1543d0F83489e82A1344DF6827B23d541F235A50",
            "symbol": "ATH (AIgatha Token)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x17052d51E954592C1046320c2371AbaB6C73Ef10",
            "symbol": "ATH (Athenian Warrior Token)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x78B7FADA55A64dD895D8c8c35779DD8b67fA8a05",
            "symbol": "ATL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9B11EFcAAA1890f6eE52C6bB7CF8153aC5d74139",
            "symbol": "ATM",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x97AEB5066E1A590e868b511457BEb6FE99d329F5",
            "symbol": "ATMI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2dAEE1AA61D60A252DC80564499A69802853583A",
            "symbol": "ATS",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x887834d3b8d450b6bab109c252df3da286d73ce4",
            "symbol": "ATT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6339784d9478da43106a429196772a029c2f177d",
            "symbol": "ATTN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1a0f2ab46ec630f9fd638029027b552afa64b94c",
            "symbol": "ATX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc12d099be31567add4e4e4d0d45691c3f58f5663",
            "symbol": "AUC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcdcfc0f66c522fd086a1b725ea3c0eeb9f9e8814",
            "symbol": "AURA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x622dFfCc4e83C64ba959530A5a5580687a57581b",
            "symbol": "AUTO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xeD247980396B10169BB1d36f6e278eD16700a60f",
            "symbol": "AVA",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x0d88ed6e74bbfd96b831231638b66c05571e824f",
            "symbol": "AVT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcd4b4b0f3284a33ac49c67961ec6e111708318cf",
            "symbol": "AX1",
            "decimal": 5,
            "type": "default"
        },
        {
            "address": "0xC39E626A04C5971D770e319760D7926502975e47",
            "symbol": "AXPR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5d51fcced3114a8bb5e90cdd0f9d682bcbcc5393",
            "symbol": "B2BX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x998b3b82bc9dba173990be7afb772788b5acb8bd",
            "symbol": "BANCA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf87f0d9153fea549c728ad61cb801595a68b73de",
            "symbol": "BANX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2a05d22db079bc40c2f77a1d1ff703a56e631cc1",
            "symbol": "BAS",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x0D8775F648430679A709E98d2b0Cb6250d2887EF",
            "symbol": "BAT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9a0242b7a33dacbe40edb927834f96eb39f8fbcb",
            "symbol": "BAX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe7D3e4413E29ae35B0893140F4500965c74365e5",
            "symbol": "BBC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x37d40510a2f5bc98aa7a0f7bf4b3453bcfb90ac1",
            "symbol": "BBI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4a6058666cf1057eaC3CD3A5a614620547559fc9",
            "symbol": "BBK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x35a69642857083ba2f30bfab735dacc7f0bac969",
            "symbol": "BBN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x84f7c44b6fed1080f647e354d552595be2cc602f",
            "symbol": "BBO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2ecb13a8c458c379c4d9a7259e202de03c8f3d19",
            "symbol": "BC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1f41e42d0a9e3c0dd3ba15b527342783b43200a9",
            "symbol": "BCAP",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x7367a68039d4704f30bfbf6d948020c3b07dfc59",
            "symbol": "BCBC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1e797Ce986C3CFF4472F7D38d5C4aba55DfEFE40",
            "symbol": "BCDN",
            "decimal": 15,
            "type": "default"
        },
        {
            "address": "0xacfa209fb73bf3dd5bbfb1101b9bc999c49062a5",
            "symbol": "BCDT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbc1234552EBea32B5121190356bBa6D3Bb225bb5",
            "symbol": "BCL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1c4481750daa5Ff521A2a7490d9981eD46465Dbd",
            "symbol": "BCPT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1014613e2b3cbc4d575054d4982e580d9b99d7b1",
            "symbol": "BCV",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1961B3331969eD52770751fC718ef530838b6dEE",
            "symbol": "BDG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4D8fc1453a0F359e99c9675954e656D80d996FbF",
            "symbol": "BEE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x74C1E4b8caE59269ec1D85D3D4F324396048F4ac",
            "symbol": "BeerCoin",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x6aeb95f06cda84ca345c2de0f3b7f96923a44f4c",
            "symbol": "BERRY",
            "decimal": 14,
            "type": "default"
        },
        {
            "address": "0x8aA33A7899FCC8eA5fBe6A608A109c3893A1B8b2",
            "symbol": "BET",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x14C926F2290044B647e1Bf2072e67B495eff1905",
            "symbol": "BETHER",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x763186eb8d4856d536ed4478302971214febc6a9",
            "symbol": "BETR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xee74110fb5a1007b06282e0de5d73a61bf41d9cd",
            "symbol": "BHPC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfe5d908c9ad85f651185daa6a4770726e2b27d09",
            "symbol": "BHR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x08b4c866aE9D1bE56a06e0C302054B4FFe067b43",
            "symbol": "BITCAR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf3d29fb98d2dc5e78c87198deef99377345fd6f1",
            "symbol": "BITPARK",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xb3104b4b9da82025e8b9f8fb28b3553ce2f67069",
            "symbol": "BIX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb2bfeb70b903f1baac7f2ba2c62934c7e5b974c4",
            "symbol": "BKB",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x3cf9e0c385a5abec9fd2a71790aa344c4e8e3570",
            "symbol": "BKRx",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x45245bc59219eeaaf6cd3f382e078a461ff9de7b",
            "symbol": "BKX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xca29db4221c111888a7e80b12eac8a266da3ee0d",
            "symbol": "BLN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x107c4504cd79C5d2696Ea0030a8dD4e92601B82e",
            "symbol": "BLT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x539efe69bcdd21a83efd9122571a64cc25e0282b",
            "symbol": "BLUE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xce59d29b09aae565feeef8e52f47c3cd5368c663",
            "symbol": "BLX (Bullion)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE5a7c12972f3bbFe70ed29521C8949b8Af6a0970",
            "symbol": "BLX (Iconomi)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5732046A883704404F284Ce41FfADd5b007FD668",
            "symbol": "BLZ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdf6ef343350780bf8c3410bf062e0c015b1dd671",
            "symbol": "BMC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf028adee51533b1b47beaa890feb54a457f51e89",
            "symbol": "BMT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x986EE2B944c42D017F52Af21c4c69B84DBeA35d8",
            "symbol": "BMX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb8c77482e45f1f44de1745f52c74426c631bdd52",
            "symbol": "BNB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdD6Bf56CA2ada24c683FAC50E37783e55B57AF9F",
            "symbol": "BNC",
            "decimal": 12,
            "type": "default"
        },
        {
            "address": "0xef51c9377feb29856e61625caf9390bd0b67ea18",
            "symbol": "BNC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xdA2C424Fc98c741c2d4ef2f42897CEfed897CA75",
            "symbol": "BNFT",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0xda80b20038bdf968c7307bb5907a469482cf6251",
            "symbol": "BNN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C",
            "symbol": "BNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd2d6158683aee4cc838067727209a0aaf4359de3",
            "symbol": "BNTY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xDF347911910b6c9A4286bA8E2EE5ea4a39eB2134",
            "symbol": "BOB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xCc34366E3842cA1BD36c1f324d15257960fCC801",
            "symbol": "BON",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7f1e2c7d6a69bf34824d72c53b4550e895c0d8c2",
            "symbol": "BOP",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xC2C63F23ec5E97efbD7565dF9Ec764FDc7d4e91d",
            "symbol": "BOU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x139d9397274bb9e2c29a9aa8aa0b5874d30d62e3",
            "symbol": "BOUTS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x63f584fa56e60e4d0fe8802b27c7e6e3b33e007f",
            "symbol": "BOX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe1A178B681BD05964d3e3Ed33AE731577d9d96dD",
            "symbol": "BOX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x780116d91e5592e58a3b3c76a351571b39abcec6",
            "symbol": "BOXX",
            "decimal": 15,
            "type": "default"
        },
        {
            "address": "0x327682779bAB2BF4d1337e8974ab9dE8275A7Ca8",
            "symbol": "BPT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9E77D5a1251b6F7D456722A6eaC6D2d5980bd891",
            "symbol": "BRAT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x558ec3152e2eb2174905cd19aea4e34a23de9ad6",
            "symbol": "BRD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf26ef5e0545384b7dcc0f297f2674189586830df",
            "symbol": "BSDC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x509A38b7a1cC0dcd83Aa9d06214663D9eC7c7F4a",
            "symbol": "BST",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x02725836ebf3ecdb1cdf1c7b02fcbbfaa2736af8",
            "symbol": "BTCA",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x0886949c1b8C412860c4264Ceb8083d1365e86CF",
            "symbol": "BTCE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x5acD19b9c91e596b1f062f18e3D02da7eD8D1e50",
            "symbol": "BTCL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x87f5e8c3425218837f3cb67db941af0c01323e56",
            "symbol": "BTCONE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6aac8cb9861e42bf8259f5abdc6ae3ae89909e11",
            "symbol": "BTCR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x73dd069c299a5d691e9836243bcaec9c8c1d8734",
            "symbol": "BTE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xfad572db566e5234ac9fc3d570c4edc0050eaa92",
            "symbol": "BTH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdb8646f5b487b5dd979fac618350e85018f557d4",
            "symbol": "BTK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2accaB9cb7a48c3E82286F0b2f8798D201F4eC3f",
            "symbol": "BTL (Battle)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x92685E93956537c25Bb75D5d47fca4266dd628B8",
            "symbol": "BTL (Bitlle)",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0xcb97e65f07da24d46bcdd078ebebd7c6e6e3d750",
            "symbol": "BTM",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x36905fc93280f52362a1cbab151f25dc46742fb5",
            "symbol": "BTO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x16B0E62aC13a2fAeD36D18bce2356d25Ab3CfAD3",
            "symbol": "BTQ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x499a6b77bc25c26bcf8265e2102b1b3dd1617024",
            "symbol": "BTR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x03c780cd554598592b97b7256ddaad759945b125",
            "symbol": "BTRN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x080aa07e2c7185150d7e4da98838a8d2feac3dfc",
            "symbol": "BTT (Bitether)",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xFA456Cf55250A839088b27EE32A424d7DAcB54Ff",
            "symbol": "BTT (Blocktrade)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb683D83a532e2Cb7DFa5275eED3698436371cc9f",
            "symbol": "BTU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE5f867dE1EA81346df5181b8b48DD6B0BB3357B0",
            "symbol": "BTZ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xca3c18a65b802ec267f8f4802545e7f53d24c75e",
            "symbol": "BUC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbd168cbf9d3a375b38dc51a202b5e8a4e52069ed",
            "symbol": "BWX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4375e7ad8a01b8ec3ed041399f62d9cd120e0063",
            "symbol": "BZ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe1aee98495365fc179699c1bb3e761fa716bee62",
            "symbol": "BZNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3839d8ba312751aa0248fed6a8bacb84308e20ed",
            "symbol": "Bez",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x26E75307Fc0C021472fEb8F727839531F112f317",
            "symbol": "C20",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd42debE4eDc92Bd5a3FBb4243e1ecCf6d63A4A5d",
            "symbol": "C8",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7d4b8Cce0591C9044a22ee543533b72E976E36C3",
            "symbol": "CAG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1d462414fe14cf489c7A21CaC78509f4bF8CD7c0",
            "symbol": "CAN",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x04f2e7221fdb1b52a68169b25793e51478ff0329",
            "symbol": "CAPP",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x423e4322cdda29156b49a17dfbd2acc4b280600d",
            "symbol": "CAR",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x4d9e23a3842fe7eb7682b9725cf6c507c424a41b",
            "symbol": "CAR (Carblock)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa517a46baad6b054a76bd19c46844f717fe69fea",
            "symbol": "CARB",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x954b890704693af242613edef1b603825afcd708",
            "symbol": "CARD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbf18f246b9301f231e9561b35a3879769bb46375",
            "symbol": "CARE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe8780B48bdb05F928697A5e8155f672ED91462F7",
            "symbol": "CAS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1234567461d3f8db7496581774bd869c83d51c93",
            "symbol": "CAT (BitClave)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x56ba2Ee7890461f463F7be02aAC3099f6d5811A8",
            "symbol": "CAT (Blockcat)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x68e14bb5A45B9681327E16E528084B9d962C1a39",
            "symbol": "CATs (BitClave)_Old",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x26db5439f651caf491a87d48799da81f191bdb6b",
            "symbol": "CBC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x076C97e1c869072eE22f8c91978C99B4bcB02591",
            "symbol": "CBT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc166038705FFBAb3794185b3a9D925632A1DF37D",
            "symbol": "CC3",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x28577A6d31559bd265Ce3ADB62d0458550F7b8a7",
            "symbol": "CCC (CryptoCrashCourse)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbe11eeb186e624b8f26a5045575a1340e4054552",
            "symbol": "CCC (ICONOMI)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x378903a03fb2c3ac76bb52773e3ce11340377a32",
            "symbol": "CCCX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd348e07a2806505b856123045d27aeed90924b50",
            "symbol": "CCLC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x679badc551626e01b23ceecefbc9b877ea18fc46",
            "symbol": "CCO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x315ce59fafd3a8d562b7ec1c8542382d2710b06c",
            "symbol": "CCS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x336f646f87d9f6bc6ed42dd46e8b3fd9dbd15c22",
            "symbol": "CCT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8a95ca448A52C0ADf0054bB3402dC5e09CD6B232",
            "symbol": "CDL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x177d39AC676ED1C67A2b268AD7F1E58826E5B0af",
            "symbol": "CDT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2cb101d7da0ebaa57d3f2fef46d7ffb7bb64592b",
            "symbol": "CDX (Carbon Dollar X)",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x6fFF3806Bbac52A20e0d79BC538d527f6a22c96b",
            "symbol": "CDX (Network)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb056c38f6b7dc4064367403e26424cd2c60655e1",
            "symbol": "CEEK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1122b6a0e00dce0563082b6e2953f3a943855c1f",
            "symbol": "CENNZ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf660ca1e228e7be1fa8b4f5583145e31147fb577",
            "symbol": "CET",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5dff89a2caa4d76bc286f74d67bd718eb834da61",
            "symbol": "CFC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x12FEF5e57bF45873Cd9B62E9DBd7BFb99e32D73e",
            "symbol": "CFI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6956983f8b3ce173b4ab84361aa0ad52f38d936f",
            "symbol": "CFTY",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf3db7560e820834658b590c96234c333cd3d5e5e",
            "symbol": "CHP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xba9d4199fab4f26efe3551d490e3821486f135ba",
            "symbol": "CHSB",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1460a58096d80a50a2f1f956dda497611fa4f165",
            "symbol": "CHX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3abdff32f76b42e7635bdb7e425f0231a5f3ab17",
            "symbol": "CJT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x06012c8cf97bead5deae237070f9587f8e7a266d",
            "symbol": "CK",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xe81d72d14b1516e68ac3190a46c93302cc8ed60f",
            "symbol": "CL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb1c1cb8c7c1992dba24e628bf7d38e71dad46aeb",
            "symbol": "CLB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3dc9a42fa7afe57be03c58fd7f4411b1e466c508",
            "symbol": "CLL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4162178B78D6985480A308B2190EE5517460406D",
            "symbol": "CLN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7fce2856899a6806eeef70807985fc7554c66340",
            "symbol": "CLP",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x3edd235c3e840c1f29286b2e39370a255c7b6fdb",
            "symbol": "CMBT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x7e667525521cF61352e2E01b50FaaaE7Df39749a",
            "symbol": "CMC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x47bc01597798dcd7506dcca36ac4302fc93a8cfb",
            "symbol": "CMCT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf85fEea2FdD81d51177F6b8F35F0e6734Ce45F5F",
            "symbol": "CMT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xEBf2F9E8De960f64ec0fDCDa6Cb282423133347B",
            "symbol": "CNB",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xd4c435f5b09f855c3317c8524cb1f586e42795fa",
            "symbol": "CND",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8713d26637cf49e1b6b4a7ce57106aabc9325343",
            "symbol": "CNN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB4b1D2C217EC0776584CE08D3DD98F90EDedA44b",
            "symbol": "CO2",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x574b36bced443338875d171cc377e691f7d4f887",
            "symbol": "CO2Bit",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb2f7eb1f2c37645be61d73953035360e768d81e6",
            "symbol": "COB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3136eF851592aCf49CA4C825131E364170FA32b3",
            "symbol": "COFI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0c91b015aba6f7b4738dcd36e7410138b29adc29",
            "symbol": "COIL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x5e8f855966d638135a968861e80dda722291b06d",
            "symbol": "COIN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc4bcd64cb216d49fd3c643a32762f34626b45a1a",
            "symbol": "COSM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x65292eeadf1426cd2df1c4793a3d7519f253913b",
            "symbol": "COSS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9e96604445ec19ffed9a5e8dd7b50a29c899a10c",
            "symbol": "COSS (coss.io)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE2FB6529EF566a080e6d23dE0bd351311087D567",
            "symbol": "COV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0Ebb614204E47c09B6C3FeB9AAeCad8EE060E23E",
            "symbol": "CPAY",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xfAE4Ee59CDd86e3Be9e8b90b53AA866327D7c090",
            "symbol": "CPC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb787d4eac8899730bb8c57fc3c998c49c5244ec0",
            "symbol": "CPEX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x7064aAb39A0Fcf7221c3396719D0917a65E35515",
            "symbol": "CPOLLO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x88d50b466be55222019d71f9e8fae17f5f45fca1",
            "symbol": "CPT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf44745fbd41f6a1ba151df190db0564c5fcc4410",
            "symbol": "CPY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7f585b9130c64e9e9f470b618a7badd03d79ca7e",
            "symbol": "CR7",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xAef38fBFBF932D1AeF3B808Bc8fBd8Cd8E1f8BC5",
            "symbol": "CRB",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x2cf618c19041d9db330d8222b860a624021f30fb",
            "symbol": "CRBT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf41e5fbc2f6aac200dd8619e121ce1f05d150077",
            "symbol": "CRC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x672a1AD4f667FB18A333Af13667aa0Af1F5b5bDD",
            "symbol": "CRED",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4e0603e2a27a30480e5e3a4fe548e29ef12f64be",
            "symbol": "CREDO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf49cdd50ad408d387d611f88a647179c3de3492b",
            "symbol": "CRGO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9238bfb781a55eacc3cf05f7df94038c198cd9b9",
            "symbol": "CRMT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x80a7e048f37a50500351c204cb407766fa3bae7f",
            "symbol": "CRPT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xF0da1186a4977226b9135d0613ee72e229EC3F4d",
            "symbol": "CRT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x46b9ad944d1059450da1163511069c718f699d31",
            "symbol": "CS",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x29d75277ac7f0335b2165d0895e8725cbf658d73",
            "symbol": "CSNO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xbb49a51ee5a66ca3a8cbe529379ba44ba67e6771",
            "symbol": "CST",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4545750F39aF6Be4F237B6869D4EccA928Fd5A85",
            "symbol": "CTF",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc87c5dd86a3d567ff28701886fb0745aaa898da4",
            "symbol": "CTG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9e7d29bd499b6c7da2a5b2eafcf4a39d3bd845d1",
            "symbol": "CTGC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbf4cfd7d1edeeea5f6600827411b41a21eb08abd",
            "symbol": "CTL",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x96A65609a7B84E8842732DEB08f56C3E21aC6f8a",
            "symbol": "CTR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE3Fa177AcecfB86721Cf6f9f4206bd3Bd672D7d5",
            "symbol": "CTT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x662aBcAd0b7f345AB7FfB1b1fbb9Df7894f18e66",
            "symbol": "CTX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xea11755ae41d889ceec39a63e6ff75a02bc1c00d",
            "symbol": "CTXC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdA6cb58A0D0C01610a29c5A65c303e13e885887C",
            "symbol": "cV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x41e5560054824ea6b0732e656e3ad64e20e94e45",
            "symbol": "CVC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xbe428c3867f05dea2a89fc76a102b544eac7f772",
            "symbol": "CVT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2134057c0b461f898d375cead652acae62b59541",
            "symbol": "CXC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb6EE9668771a79be7967ee29a63D4184F8097143",
            "symbol": "CXO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3f06B5D78406cD97bdf10f5C420B241D32759c80",
            "symbol": "CYFM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x78c292d1445e6b9558bf42e8bc369271ded062ea",
            "symbol": "CYMT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x0223fc70574214f65813fe336d870ac47e147fae",
            "symbol": "CZR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE4c94d45f7Aef7018a5D66f44aF780ec6023378e",
            "symbol": "CryptoCarbon",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x05c3617cbf1304b9260aa61ec960f115d67becea",
            "symbol": "Cubrix",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdab0C31BF34C897Fb0Fe90D12EC9401caf5c36Ec",
            "symbol": "DAB",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xa31108e5bab5494560db34c95492658af239357c",
            "symbol": "DACS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfb2f26f266fb2805a387230f2aa0a331b4d96fba",
            "symbol": "DADI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x89d24A6b4CcB1B6fAA2625fE562bDD9a23260359",
            "symbol": "DAI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x07d9e49ea402194bf48a8276dafb16e4ed633317",
            "symbol": "DALC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x9B70740e708a083C6fF38Df52297020f5DfAa5EE",
            "symbol": "DAN",
            "decimal": 10,
            "type": "default"
        },
        {
            "address": "0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413",
            "symbol": "DAO",
            "decimal": 16,
            "type": "default"
        },
        {
            "address": "0x81c9151de0c8bafcd325a57e3db5a5df1cebf79c",
            "symbol": "DAT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1b5f21ee98eed48d292e8e2d3ed82b40a9728a22",
            "symbol": "DATABroker",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0cf0ee63788a0849fe5297f3407f701e122cc023",
            "symbol": "DATACoin",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xabbbb6447b68ffd6141da77c18c7b5876ed6c5ab",
            "symbol": "DATx",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd82Df0ABD3f51425Eb15ef7580fDA55727875f14",
            "symbol": "DAV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0b4bdc478791897274652dc15ef5c135cae61e60",
            "symbol": "DAX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x61725f3db4004afe014745b21dab1e1677cc328b",
            "symbol": "DAXT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9b68bfae21df5a510931a262cecf63f41338f264",
            "symbol": "DBET",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x386Faa4703a34a7Fdb19Bec2e14Fd427C9638416",
            "symbol": "DCA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xffa93aacf49297d51e211817452839052fdfb961",
            "symbol": "DCC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x399A0e6FbEb3d74c85357439f4c8AeD9678a5cbF",
            "symbol": "DCL",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x08d32b0da63e2C3bcF8019c9c5d849d7a9d791e6",
            "symbol": "DCN",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xcC4eF9EEAF656aC1a2Ab886743E98e97E090ed38",
            "symbol": "DDF",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x151202C9c18e495656f372281F493EB7698961D5",
            "symbol": "DEB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xde1e0ae6101b46520cf66fdc0b1059c5cc3d106c",
            "symbol": "DELTA",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x3597bfd533a99c9aa083587b074434e61eb0a258",
            "symbol": "DENT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x7cF271966F36343Bf0150F25E5364f7961c58201",
            "symbol": "DEPO",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x89cbeac5e8a13f0ebb4c74fadfc69be81a501106",
            "symbol": "DEPO (Depository Network)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdd94De9cFE063577051A5eb7465D08317d8808B6",
            "symbol": "Devcon2 Token",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x20e94867794dba030ee287f1406e100d03c84cd3",
            "symbol": "DEW",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE0B7927c4aF23765Cb51314A0E0521A9645F0E2A",
            "symbol": "DGD",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0xf6cFe53d6FEbaEEA051f400ff5fc14F0cBBDacA1",
            "symbol": "DGPT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6aEDbF8dFF31437220dF351950Ba2a3362168d1b",
            "symbol": "DGS",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1c83501478f1320977047008496dacbd60bb15ef",
            "symbol": "DGTX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4f3afec4e5a3f2a6a1a411def7d7dfe50ee057bf",
            "symbol": "DGX",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x55b9a11c2e8351b4Ffc7b11561148bfaC9977855",
            "symbol": "DGX1",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x2e071D2966Aa7D8dECB1005885bA1977D6038A65",
            "symbol": "DICE",
            "decimal": 16,
            "type": "default"
        },
        {
            "address": "0xc719d010B63E5bbF2C0551872CD5316ED26AcD83",
            "symbol": "DIP (Decentralized Insurance Protocol)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf14922001a2fb8541a433905437ae954419c2439",
            "symbol": "DIT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x13f11C9905A08ca76e3e853bE63D4f0944326C72",
            "symbol": "DIVX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x07e3c70653548b04f0a75970c1f81b4cbbfb606f",
            "symbol": "DLT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2ccbFF3A042c68716Ed2a2Cb0c544A9f1d1935E1",
            "symbol": "DMT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x82b0e50478eeafde392d45d1259ed1071b6fda81",
            "symbol": "DNA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0abdace70d3790235af448c88547603b945604ea",
            "symbol": "DNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE43E2041dc3786e166961eD9484a5539033d10fB",
            "symbol": "DNX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe5dada80aa6477e85d09747f2842f7993d0df71c",
            "symbol": "DOCK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x906b3f8b7845840188eab53c3f5ad348a787752f",
            "symbol": "DOR",
            "decimal": 15,
            "type": "default"
        },
        {
            "address": "0xac3211a5025414af2866ff09c23fc18bc97e79b1",
            "symbol": "DOV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x76974c7b79dc8a6a109fd71fd7ceb9e40eff5382",
            "symbol": "DOW",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x01b3Ec4aAe1B8729529BEB4965F27d008788B0EB",
            "symbol": "DPP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x419c4db4b9e25d6db2ad9691ccb832c8d9fda05e",
            "symbol": "DRGN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4672bad527107471cb5067a887f4656d585a8a31",
            "symbol": "DROP (dropil)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3c75226555FC496168d48B88DF83B95F16771F37",
            "symbol": "DROP (droplex)",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x621d78f2ef2fd937bfca696cabaf9a779f59b3ed",
            "symbol": "DRP",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x2799d90c6d44cb9aa5fbc377177f16c33e056b82",
            "symbol": "DRP (Dripcoin)",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xe30e02f049957e2a5907589e06ba646fb2c321ba",
            "symbol": "DRPU",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x9af4f26941677c706cfecf6d3379ff01bb85d5ab",
            "symbol": "DRT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x62d4c04644314f35868ba4c65cc27a77681de7a9",
            "symbol": "DRVH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x03e3f0c25965f13DbbC58246738C183E27b26a56",
            "symbol": "DSCP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5adc961D6AC3f7062D2eA45FEFB8D8167d44b190",
            "symbol": "DTH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd234bf2410a0009df9c3c63b610c09738f18ccd7",
            "symbol": "DTR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xc20464e0c373486d2b3335576e83a218b1618a5e",
            "symbol": "DTRC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf9F7c29CFdf19FCf1f2AA6B84aA367Bcf1bD1676",
            "symbol": "DTT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x765f0c16d1ddc279295c1a7c24b0883f62d33f75",
            "symbol": "DTX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x82fdedfB7635441aA5A92791D001fA7388da8025",
            "symbol": "DTx",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xEd7fEA78C393cF7B17B152A8c2D0CD97aC31790B",
            "symbol": "DUBI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8db54ca569d3019a2ba126d03c37c44b5ef81ef6",
            "symbol": "DXT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xce5c603c78d047ef43032e96b5b785324f753a4f",
            "symbol": "E4ROW",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x994f0dffdbae0bbf09b652d6f11a493fd33f42b9",
            "symbol": "EAGLE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x900b4449236a7bb26b286601dd14d2bde7a6ac6c",
            "symbol": "EARTH",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x31f3d9d1bece0c033ff78fa6da60a6048f3e13c5",
            "symbol": "EBC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xafc39788c51f0c1ff7b55317f3e70299e521fff6",
            "symbol": "eBCH",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xa578acc0cb7875781b7880903f4594d13cfa8b98",
            "symbol": "ECN",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x17F93475d2A978f527c3f7c44aBf44AdfBa60D5C",
            "symbol": "ECO2",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x171d750d42d661b62c277a6b486adb82348c3eca",
            "symbol": "ECOM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8869b1f9bc8b246a4d7220f834e56ddfdd8255e7",
            "symbol": "ECP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfa1de2ee97e4c10c94c91cb2b5062b89fb140b82",
            "symbol": "EDC",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x08711D3B02C8758F2FB3ab4e80228418a7F8e39c",
            "symbol": "EDG",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xced4e93198734ddaff8492d525bd258d49eb388e",
            "symbol": "EDO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc528c28FEC0A90C083328BC45f587eE215760A0F",
            "symbol": "EDR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2a22e5cca00a3d63308fa39f29202eb1b39eef52",
            "symbol": "EDU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xeb7c20027172e5d143fb030d50f91cece2d1485d",
            "symbol": "eBTC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xb53a96bcbdd9cf78dff20bab6c2be7baec8f00f8",
            "symbol": "eGAS",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x8e1b448EC7aDFc7Fa35FC2e885678bD323176E34",
            "symbol": "EGT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5dbac24e98e2a4f43adc0dc82af403fca063ce2c",
            "symbol": "EGT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf9F0FC7167c311Dd2F1e21E9204F87EBA9012fB2",
            "symbol": "EHT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xa6a840e50bcaa50da017b91a0d86b8b2d41156ee",
            "symbol": "EKO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbab165df9455aa0f2aed1f2565520b91ddadb4c8",
            "symbol": "EKT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xd49ff13661451313ca1553fd6954bd1d9b6e02b9",
            "symbol": "ELEC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbf2179859fc6d5bee9bf9158632dc51678a4100e",
            "symbol": "ELF",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc8C6A31A4A806d3710A7B38b7B296D2fABCCDBA8",
            "symbol": "ELIX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x44197a4c44d6a059297caf6be4f7e172bd56caaf",
            "symbol": "ELTCOIN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xa95592DCFfA3C080B4B40E459c5f5692F67DB7F8",
            "symbol": "ELY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb67b88a25708a35ae7c2d736d398d268ce4f7f83",
            "symbol": "EMON",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x95daaab98046846bf4b2853e23cba236fa394a31",
            "symbol": "EMONT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x9501BFc48897DCEEadf73113EF635d2fF7ee4B97",
            "symbol": "EMT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB802b24E0637c2B87D2E8b7784C055BBE921011a",
            "symbol": "EMV",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x039f5050de4908f9b5ddf40a4f3aa3f329086387",
            "symbol": "ENC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf0ee6b27b759c9893ce4f094b49ad28fd15a23e4",
            "symbol": "ENG",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xF629cBd94d3791C9250152BD8dfBDF380E2a3B9c",
            "symbol": "ENJ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5BC7e5f0Ab8b2E10D2D0a3F21739FCe62459aeF3",
            "symbol": "ENTRP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x86fa049857e0209aa7d9e616f7eb3b3b78ecfdb0",
            "symbol": "EOS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7e9e431a0b8c4d532c745b1043c7fa29a48d4fba",
            "symbol": "eosDAC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x35BAA72038F127f9f8C8f9B491049f64f377914d",
            "symbol": "EPX",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x50ee674689d75c0f88e8f83cfe8c4b69e8fd590d",
            "symbol": "EPY",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x47dd62d4d075dead71d0e00299fc56a2d747bebb",
            "symbol": "EQL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x74ceda77281b339142a36817fa5f9e29412bab85",
            "symbol": "ERO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x92A5B04D0ED5D94D7a193d1d334D3D16996f4E13",
            "symbol": "ERT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe8a1df958be379045e2b46a31a98b93a2ecdfded",
            "symbol": "ESZ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1b9743f556d65e757c4c650b4555baf354cb8bd3",
            "symbol": "ETBS",
            "decimal": 12,
            "type": "default"
        },
        {
            "address": "0xdd74a7a3769fa72561b3a69e65968f49748c690c",
            "symbol": "ETCH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x28c8d01ff633ea9cd8fc6a451d7457889e698de6",
            "symbol": "ETG",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x3a26746Ddb79B1B8e4450e3F4FFE3285A307387E",
            "symbol": "ETHB",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xdbfb423e9bbf16294388e07696a5120e4ceba0c5",
            "symbol": "ETHD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5Af2Be193a6ABCa9c8817001F45744777Db30756",
            "symbol": "ETHOS",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x3c4a3ffd813a107febd57b2f01bc344264d90fde",
            "symbol": "ETK",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x6927C69fb4daf2043fbB1Cb7b86c5661416bea29",
            "symbol": "ETR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdb25f211ab05b1c97d595516f45794528a807ad8",
            "symbol": "EURS",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0xabdf147870235fcfc34153828c769a70b3fae01f",
            "symbol": "EURT",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xb62d18dea74045e822352ce4b3ee77319dc5ff2f",
            "symbol": "EVC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x923108a439C4e8C2315c4f6521E5cE95B44e9B4c",
            "symbol": "EVE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x68909e586eeac8f47315e84b4c9788dd54ef65bb",
            "symbol": "EVN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd780Ae2Bf04cD96E577D3D014762f831d97129d0",
            "symbol": "EVN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf3db5fa2c66b7af3eb0c0b782510816cbe4813b8",
            "symbol": "EVX",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x444997b7e7fC830E20089afea3078cd518fCF2A2",
            "symbol": "EWO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x00c4b398500645eb5da00a1a379a88b11683ba01",
            "symbol": "EXC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc98e0639c6d2ec037a615341c369666b110e80e5",
            "symbol": "EXMR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xe469c4473af82217b30cf17b10bcdb6c8c796e75",
            "symbol": "EXRN",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x5c743a35e903f6c584514ec617acee0611cf44f3",
            "symbol": "EXY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5e6016ae7d7c49d347dcf834860b9f3ee282812b",
            "symbol": "EZT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xb67734521eAbBE9C773729dB73E16CC2dfb20A58",
            "symbol": "E₹",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x1ccaa0f2a7210d76e1fdec740d5f323e2e1b1672",
            "symbol": "FACE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x190e569bE071F40c704e15825F285481CB74B6cC",
            "symbol": "FAM",
            "decimal": 12,
            "type": "default"
        },
        {
            "address": "0x90162f41886c0946d09999736f1c15c8a105a421",
            "symbol": "FAN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7dcb3b2356c822d3577d4d060d0d5d78c860488c",
            "symbol": "FANX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x23352036e911a22cfc692b5e2e196692658aded9",
            "symbol": "FDZ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x52fb36c83ad33c1824912fc81071ca5eeb8ab390",
            "symbol": "FID",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x009e864923b49263c7F10D19B7f8Ab7a9A5AAd33",
            "symbol": "FKX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf04a8ac553FceDB5BA99A64799155826C136b0Be",
            "symbol": "FLIXX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x04cC783b450b8D11F3C7d00DD03fDF7FB51fE9F2",
            "symbol": "FLMC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x049399a6b048d52971f7d122ae21a1532722285f",
            "symbol": "FLOT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3a1Bda28AdB5B0a812a7CF10A1950c920F79BcD3",
            "symbol": "FLP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9aeFBE0b3C3ba9Eab262CB9856E8157AB7648e09",
            "symbol": "FLR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x954b5de09a55e59755acbda29e1eb74a45d30175",
            "symbol": "FLUZ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x70b147e01e9285e7ce68b9ba437fe3a9190e756a",
            "symbol": "FLX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4df47b4969b2911c966506e3592c41389493953b",
            "symbol": "FND",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0707681f344deb24184037fc0228856f2137b02e",
            "symbol": "FNKOS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbd4b60a138b3fce3584ea01f50c0908c18f9677a",
            "symbol": "FNTB",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x2a093BcF0C98Ef744Bb6F69D74f2F85605324290",
            "symbol": "FOOD",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x4270bb238f6dd8b1c3ca01f96ca65b2647c06d3c",
            "symbol": "FOTA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0ABeFb7611Cb3A01EA3FaD85f33C3C934F8e2cF4",
            "symbol": "FRD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x17e67d1cb4e349b9ca4bc3e17c7df2a397a7bb64",
            "symbol": "FREC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x48df4e0296f908ceab0428a5182d19b31fc037d6",
            "symbol": "FRV",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xd0352a019e9ab9d757776f532377aaebd36fd541",
            "symbol": "FSN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x78a73B6CBc5D183CE56e786f6e905CaDEC63547B",
            "symbol": "FT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe6f74dcfa0e20883008d8c16b6d9a329189d0c30",
            "symbol": "FTC",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x943ed852dadb5c3938ecdc6883718df8142de4c8",
            "symbol": "FTI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2023DCf7c438c8C8C0B0F28dBaE15520B4f3Ee20",
            "symbol": "FTR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2AEC18c5500f21359CE1BEA5Dc1777344dF4C0Dc",
            "symbol": "FTT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd559f20296ff4895da39b5bd9add54b442596a61",
            "symbol": "FTX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x41875c2332b0877cdfaa699b641402b7d4642c32",
            "symbol": "FTXT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x65be44c747988fbf606207698c944df4442efe19",
            "symbol": "FUCK",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0xEA38eAa3C86c8F9B751533Ba2E562deb9acDED40",
            "symbol": "FUEL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x419D0d8BdD9aF5e606Ae2232ed285Aff190E711b",
            "symbol": "FUN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1829aa045e21e0d59580024a951db48096e01782",
            "symbol": "FXT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x88FCFBc22C6d3dBaa25aF478C578978339BDe77a",
            "symbol": "FYN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8f0921f30555624143d427b340b1156914882c10",
            "symbol": "FYP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf67451dc8421f0e0afeb52faa8101034ed081ed9",
            "symbol": "GAM",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xc0EA6306F6360FE7dCAB65D16Bf1a3AF92C79Aa2",
            "symbol": "GANA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x687174f8c49ceb7729d925c3a961507ea4ac7b28",
            "symbol": "GAT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x708876f486e448ee89eb332bfbc8e593553058b9",
            "symbol": "GAVEL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7585F835ae2d522722d2684323a0ba83401f32f5",
            "symbol": "GBT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x12fCd6463E66974cF7bBC24FFC4d40d6bE458283",
            "symbol": "GBX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xdb0F69306FF8F949f258E83f6b87ee5D052d0b23",
            "symbol": "GCP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4F4f0Db4de903B88f2B1a2847971E231D54F8fd3",
            "symbol": "GEE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x24083bb30072643c3bb90b44b7285860a755e687",
            "symbol": "GELD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc7bba5b765581efb2cdd2679db5bea9ee79b201f",
            "symbol": "GEM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x543Ff227F64Aa17eA132Bf9886cAb5DB55DCAddf",
            "symbol": "GEN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8a854288a5976036a725879164ca3e91d30c6a1b",
            "symbol": "GET",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xFcD862985628b254061F7A918035B80340D045d3",
            "symbol": "GIF",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xaE4f56F072c34C0a65B3ae3E4DB797D831439D93",
            "symbol": "GIM",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x71d01db8d6a2fbea7f8d434599c237980c234e4c",
            "symbol": "GLA",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xb3Bd49E28f8F832b8d1E246106991e546c323502",
            "symbol": "GMT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6810e776880C02933D47DB1b9fc05908e5386b96",
            "symbol": "GNO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa74476443119A942dE498590Fe1f2454d7D4aC0d",
            "symbol": "GNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6ec8a24cabdc339a06a172f8223ea557055adaa5",
            "symbol": "GNX",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0xeAb43193CF0623073Ca89DB9B712796356FA7414",
            "symbol": "GOLDX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x423b5f62b328d0d6d44870f4eee316befa0b2df5",
            "symbol": "GOT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x12b19d3e2ccc14da04fae33e63652ce469b3f2fd",
            "symbol": "GRID",
            "decimal": 12,
            "type": "default"
        },
        {
            "address": "0xb444208cb0516c150178fcf9a52604bc04a1acea",
            "symbol": "GRMD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0a9A9ce600D08BF9b76F49FA4e7b38A67EBEB1E6",
            "symbol": "GROW",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x228ba514309ffdf03a81a205a6d040e429d6e80c",
            "symbol": "GSC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe530441f4f73bDB6DC2fA5aF7c3fC5fD551Ec838",
            "symbol": "GSE",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0xB70835D7822eBB9426B56543E391846C107bd32C",
            "symbol": "GTC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x025abad9e518516fdaafbdcdb9701b37fb7ef0fa",
            "symbol": "GTKT",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xc5bbae50781be1669306b9e001eff57a2957b09d",
            "symbol": "GTO",
            "decimal": 5,
            "type": "default"
        },
        {
            "address": "0xbdcfbf5c4d91abc0bc9709c7286d00063c0e6f22",
            "symbol": "GUESS",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x9847345de8b614c956146bbea549336d9c8d26b6",
            "symbol": "GULD",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf7B098298f7C69Fc14610bf71d5e02c60792894C",
            "symbol": "GUP",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x103c3A209da59d3E7C4A89307e66521e081CFDF0",
            "symbol": "GVT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x58ca3065c0f24c7c96aee8d6056b5b5decf9c2f8",
            "symbol": "GXC",
            "decimal": 10,
            "type": "default"
        },
        {
            "address": "0x22F0AF8D78851b72EE799e05F54A77001586B18A",
            "symbol": "GXVC",
            "decimal": 10,
            "type": "default"
        },
        {
            "address": "0x8C65e992297d5f092A756dEf24F4781a280198Ff",
            "symbol": "GZE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE638dc39b6aDBEE8526b5C22380b4b45dAf46d8e",
            "symbol": "GZR",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x48c1b2f3efa85fbafb2ab951bf4ba860a08cdbb7",
            "symbol": "HAND",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x5a567e28dbfa2bbd3ef13c0a01be114745349657",
            "symbol": "HAPPY",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x9002D4485b7594e3E850F0a206713B305113f69e",
            "symbol": "HAT",
            "decimal": 12,
            "type": "default"
        },
        {
            "address": "0xC011A72400E58ecD99Ee497CF89E3775d4bd732F",
            "symbol": "HAV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe2492f8d2a2618d8709ca99b1d8d75713bd84089",
            "symbol": "HB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdd6c68bb32462e01705011a4e2ad1a60740f217f",
            "symbol": "HBT",
            "decimal": 15,
            "type": "default"
        },
        {
            "address": "0xe34e1944e776f39b9252790a0527ebda647ae668",
            "symbol": "HBZ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xffe8196bc259e8dedc544d935786aa4709ec3e64",
            "symbol": "HDG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x95C4be8534d69C248C0623c4C9a7A2a001c17337",
            "symbol": "HDL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x491c9a23db85623eed455a8efdd6aba9b911c5df",
            "symbol": "HER",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xba2184520A1cC49a6159c57e61E1844E085615B6",
            "symbol": "HGT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x9bb1db1445b83213a56d90d331894b3f26218e4e",
            "symbol": "HIBT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa9240fBCAC1F0b9A6aDfB04a53c8E3B0cC1D1444",
            "symbol": "HIG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x14F37B574242D366558dB61f3335289a5035c506",
            "symbol": "HKG",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x9e6b2b11542f2bc52f3029077ace37e8fd838d7f",
            "symbol": "HKN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x88ac94d5d175130347fc95e109d77ac09dbf5ab7",
            "symbol": "HKY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x66eb65D7Ab8e9567ba0fa6E37c305956c5341574",
            "symbol": "HLX",
            "decimal": 5,
            "type": "default"
        },
        {
            "address": "0xAa0bb10CEc1fa372eb3Abc17C933FC6ba863DD9E",
            "symbol": "HMC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcbCC0F036ED4788F63FC0fEE32873d6A7487b908",
            "symbol": "HMQ",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xb45d7Bc4cEBcAB98aD09BABDF8C818B2292B672c",
            "symbol": "HODL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5B0751713b2527d7f002c0c4e2a37e1219610A6B",
            "symbol": "HORSE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6c6ee5e31d828de241282b9606c8e98ea48526e2",
            "symbol": "HOT (HoloToken)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9af839687f6c94542ac5ece2e317daae355493a1",
            "symbol": "HOT (Hydro Protocol)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x38c6a68304cdefb9bec48bbfaaba5c5b47818bb2",
            "symbol": "HPB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x554C20B7c486beeE439277b4540A434566dC4C02",
            "symbol": "HST",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6f259637dcd74c767781e37bc6133cd6a68aa161",
            "symbol": "HT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xC0Eb85285d83217CD7c891702bcbC0FC401E2D9D",
            "symbol": "HVN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xebbdf302c940c6bfd49c6b165f457fdb324649bc",
            "symbol": "HYDRO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe9ff07809ccff05dae74990e25831d0bc5cbe575",
            "symbol": "Hdp",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x84543f868ec1b1fac510d49d13c069f64cd2d5f9",
            "symbol": "Hdp.ф",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5a84969bb663fb64F6d015DcF9F622Aedc796750",
            "symbol": "ICE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x888666CA69E0f178DED6D75b5726Cee99A87D698",
            "symbol": "ICN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa33e729bf4fdeb868b534e1f20523463d9c46bee",
            "symbol": "ICO",
            "decimal": 10,
            "type": "default"
        },
        {
            "address": "0x014B50466590340D41307Cc54DCee990c8D58aa8",
            "symbol": "ICOS",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xb5a5f22694352c15b00323844ad545abb2b11028",
            "symbol": "ICX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x814cafd4782d2e728170fda68257983f03321c58",
            "symbol": "IDEA",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x5136c98a80811c3f46bdda8b5c4555cfd9f812f0",
            "symbol": "IDH",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xcc13fc627effd6e35d2d2706ea3c4d7396c610ea",
            "symbol": "IDXM",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x859a9c0b44cb7066d956a958b0b82e54c9e44b4b",
            "symbol": "iETH",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x7654915a1b82d6d2d0afc37c52af556ea8983c7e",
            "symbol": "IFT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8a88f04e0c905054d2f33b26bb3a46d7091a039a",
            "symbol": "IG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xeda8b016efa8b1161208cf041cd86972eee0f31e",
            "symbol": "IHT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x16662f73df3e79e54c6c5938b4313f92c524c120",
            "symbol": "IIC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x88AE96845e157558ef59e9Ff90E766E22E480390",
            "symbol": "IKB",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xe3831c5A982B279A198456D577cfb90424cb6340",
            "symbol": "IMC",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x22E5F62D0FA19974749faa194e3d3eF6d89c08d7",
            "symbol": "IMT (Immortal)",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x13119E34E140097a507B07a5564bDe1bC375D9e6",
            "symbol": "IMT (MoneyToken)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf8e386EDa857484f5a12e4B5DAa9984E06E73705",
            "symbol": "IND",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x24ddff6d8b8a42d835af3b440de91f3386554aa4",
            "symbol": "ING",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x48e5413b73add2434e47504E2a22d14940dBFe78",
            "symbol": "INRM",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x5b2e4a700dfbc560061e957edec8f6eeeb74a320",
            "symbol": "INS",
            "decimal": 10,
            "type": "default"
        },
        {
            "address": "0xc72fe8e3dd5bef0f9f31f259399f301272ef2a2d",
            "symbol": "INSTAR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0b76544f6c413a555f309bf76260d1e02377c02a",
            "symbol": "INT",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xece83617db208ad255ad4f45daf81e25137535bb",
            "symbol": "INV",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xa8006c4ca56f24d6836727d106349320db7fef82",
            "symbol": "INXT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xFA1a856Cfa3409CFa145Fa4e20Eb270dF3EB21ab",
            "symbol": "IOST",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6fb3e0a217407efff7ca062d46c26e5d60a14d69",
            "symbol": "IOTX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x64CdF819d3E75Ac8eC217B3496d7cE167Be42e80",
            "symbol": "IPL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x001f0aa5da15585e5b2305dbab2bac425ea71007",
            "symbol": "IPSX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0cf713b11c9b986ec40d65bd4f7fbd50f6ff2d64",
            "symbol": "IST34",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5e6b6d9abad9093fdc861ea1600eba1b355cd940",
            "symbol": "ITC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0aeF06DcCCC531e581f0440059E6FfCC206039EE",
            "symbol": "ITT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xA4eA687A2A7F29cF2dc66B39c68e4411C0D00C49",
            "symbol": "IVY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfca47962d45adfdfd1ab2d972315db4ce7ccf094",
            "symbol": "IXT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xc34b21f6f8e51cc965c2393b3ccfa3b82beb2403",
            "symbol": "IoT",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x0d262e5dc4a06a0f1c90ce79c7a60c09dfc884e4",
            "symbol": "J8T",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x884e3902C4d5cFA86de4aCE7A96AA91EbC25C0Ff",
            "symbol": "JBX",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xe2d82dc7da0e6f882e96846451f4fabcc8f90528",
            "symbol": "JC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8727c112c712c4a03371ac87a74dd6ab104af768",
            "symbol": "JET",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa5Fd1A791C4dfcaacC963D4F73c6Ae5824149eA7",
            "symbol": "JNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdb455c71c1bc2de4e80ca451184041ef32054001",
            "symbol": "JOT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xDDe12a12A6f67156e0DA672be05c374e1B0a3e57",
            "symbol": "JOY",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x773450335eD4ec3DB45aF74f34F2c85348645D39",
            "symbol": "JetCoins",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1410434b0346f5be678d0fb554e5c7ab620f8f4a",
            "symbol": "KAN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0D6DD9f68d24EC1d5fE2174f3EC8DAB52B52BaF5",
            "symbol": "KC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x72D32ac1c5E66BfC5b08806271f8eEF915545164",
            "symbol": "KEE",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x4cd988afbad37289baaf53c13e98e2bd46aaea8c",
            "symbol": "KEY (BihuKey)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4CC19356f2D37338b9802aa8E8fc58B0373296E7",
            "symbol": "KEY (SelfKey)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x27695E09149AdC738A978e9A678F99E4c39e9eb9",
            "symbol": "KICK",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x818Fc6C2Ec5986bc6E2CBf00939d90556aB12ce5",
            "symbol": "KIN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4618519de4c304f3444ffa7f812dddc2971cc688",
            "symbol": "KIND",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xdd974D5C2e2928deA5F71b9825b8b646686BD200",
            "symbol": "KNC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8e5610ab5e39d26828167640ea29823fe1dd5843",
            "symbol": "KNDC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xff5c25d2f40b47c4a37f989de933e26562ef0ac0",
            "symbol": "KNT",
            "decimal": 16,
            "type": "default"
        },
        {
            "address": "0xb5c33f965c8899d255c34cdd2a3efa8abcbb3dea",
            "symbol": "KPR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x464ebe77c293e473b48cfe96ddcf88fcf7bfdac0",
            "symbol": "KRL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdf1338FbAfe7aF1789151627B886781ba556eF9a",
            "symbol": "KUE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x241ba672574a78a3a604cdd0a94429a73a84a324",
            "symbol": "KWATT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9541FD8B9b5FA97381783783CeBF2F5fA793C262",
            "symbol": "KZN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xE50365f5D679CB98a1dd62D6F6e58e59321BcdDf",
            "symbol": "LA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfD107B473AB90e8Fbd89872144a3DC92C40Fa8C9",
            "symbol": "LALA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2f85e502a988af76f7ee6d83b7db8d6c0a823bf9",
            "symbol": "LATX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xfe5f141bf94fe84bc28ded0ab966c16b17490657",
            "symbol": "LBA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xaa19961b6b858d9f18a115f25aa1d98abc1fdba8",
            "symbol": "LCS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4a37a91eec4c97f9090ce66d21d3b3aadf1ae5ad",
            "symbol": "LCT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x05c7065d644096a4e4c3fe24af86e36de021074b",
            "symbol": "LCT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5102791ca02fc3595398400bfe0e33d7b6c82267",
            "symbol": "LDC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5b26C5D0772E5bbaC8b3182AE9a13f9BB2D03765",
            "symbol": "LEDU",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x60C24407d01782C2175D32fe7C8921ed732371D1",
            "symbol": "LEMO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x80fB784B7eD66730e8b1DBd9820aFD29931aab03",
            "symbol": "LEND",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0f4ca92660efad97a9a70cb0fe969c755439772c",
            "symbol": "LEV",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0xc798cd1c49db0e297312e4c682752668ce1db2ad",
            "symbol": "LFR",
            "decimal": 5,
            "type": "default"
        },
        {
            "address": "0x59061b6f26BB4A9cE5828A19d35CFD5A4B80F056",
            "symbol": "LGD",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x123ab195dd38b1b40510d467a6a359b201af056f",
            "symbol": "LGO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x2eb86e8fc520e0f6bb5d9af08f924fe70558ab89",
            "symbol": "LGR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xEB9951021698B42e4399f9cBb6267Aa35F82D59D",
            "symbol": "LIF",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xff18dbc487b4c2e3222d115952babfda8ba52f5f",
            "symbol": "LIFE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x02f61fd266da6e8b102d4121f5ce7b992640cf98",
            "symbol": "LIKE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x514910771af9ca656af840dff83e8264ecf986ca",
            "symbol": "LINK (Chainlink)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe2e6d4be086c6938b53b22144855eef674281639",
            "symbol": "LINK Platform",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x24A77c1F17C547105E14813e517be06b0040aa76",
            "symbol": "LIVE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x49bd2da75b1f7af1e4dfd6b1125fecde59dbec58",
            "symbol": "LKY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x63e634330A20150DbB61B15648bC73855d6CCF07",
            "symbol": "LNC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6beb418fc6e1958204ac8baddcf109b8e9694966",
            "symbol": "LNC-Linker Coin",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0947b0e6D821378805c9598291385CE7c791A6B2",
            "symbol": "LND",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5e3346444010135322268a4630d2ed5f8d09446c",
            "symbol": "LOC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9c23d67aea7b95d80942e3836bcdf7e708a747c2",
            "symbol": "LOCI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xC64500DD7B0f1794807e67802F8Abbf5F8Ffb054",
            "symbol": "LOCUS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x253c7dd074f4bacb305387f922225a4f737c08bd",
            "symbol": "LOOK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x21ae23b882a340a22282162086bc98d3e2b73018",
            "symbol": "LOOK old",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa4e8c3ec456107ea67d3075bf9e3df3a75823db0",
            "symbol": "LOOM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xEF68e7C694F40c8202821eDF525dE3782458639f",
            "symbol": "LRC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5dbe296f97b23c4a6aa6183d73e574d02ba5c719",
            "symbol": "LUC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xFB12e3CcA983B9f59D90912Fd17F8D745A8B2953",
            "symbol": "LUCK",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xa89b5934863447f6e4fc53b315a93e873bda69a3",
            "symbol": "LUM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfa05A73FfE78ef8f1a739473e462c54bae6567D9",
            "symbol": "LUN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x57ad67acf9bf015e4820fbd66ea1a21bed8852ec",
            "symbol": "LYM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3f4b726668da46f5e0e75aa5d478acec9f38210f",
            "symbol": "M-ETH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5b09a0371c1da44a8e24d36bf5deb1141a84d875",
            "symbol": "MAD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe25bCec5D3801cE3a794079BF94adF1B8cCD802D",
            "symbol": "MAN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0F5D2fB29fb7d3CFeE444a200298f468908cC942",
            "symbol": "MANA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfdcc07Ab60660de533b5Ad26e1457b565a9D59Bd",
            "symbol": "MART",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x386467f1f3ddbe832448650418311a479eecfc57",
            "symbol": "MBRS",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x93E682107d1E9defB0b5ee701C71707a4B2E46Bc",
            "symbol": "MCAP",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x138A8752093F4f9a79AaeDF48d4B9248fab93c9C",
            "symbol": "MCI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB63B606Ac810a52cCa15e44bB630fd42D8d1d83d",
            "symbol": "MCO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x51DB5Ad35C671a87207d88fC11d593AC0C8415bd",
            "symbol": "MDA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x66186008C1050627F979d464eABb258860563dbE",
            "symbol": "MDS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x814e0908b12a99fecf5bc101bb5d0b8b5cdf7d26",
            "symbol": "MDT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfd1e80508f243e64ce234ea88a5fd2827c71d4b7",
            "symbol": "MEDX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf03045a4c8077e38f3b8e2ed33b8aee69edf869f",
            "symbol": "MESH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x01f2acf2914860331c1cb1a9acecda7475e06af8",
            "symbol": "MESH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5b8d43ffde4a2982b9a5387cdf21d54ead64ac8d",
            "symbol": "MEST",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa3d58c4e56fedcae3a7c43a725aee9a71f0ece4e",
            "symbol": "MET",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfef3884b603c33ef8ed4183346e093a173c94da6",
            "symbol": "METM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6710c63432a2de02954fc0f851db07146a6c0312",
            "symbol": "MFG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xDF2C7238198Ad8B389666574f2d8bc411A4b7428",
            "symbol": "MFT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x05D412CE18F24040bB3Fa45CF2C69e506586D8e8",
            "symbol": "MFTU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x40395044ac3c0c57051906da938b54bd6557f212",
            "symbol": "MGO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x3a1237d38d0fb94513f85d61679cad7f38507242",
            "symbol": "MIC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe23cd160761f63FC3a1cF78Aa034b6cdF97d3E0C",
            "symbol": "MIT (Mainstreet)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xad8dd4c725de1d31b9e8f8d146089e9dc6882093",
            "symbol": "MIT (Mychatcoin)",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x4a527d8fc13c5203ab24ba0944f4cb14658d1db6",
            "symbol": "MITx",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
            "symbol": "MKR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7939882b54fcf0bcae6b53dec39ad6e806176442",
            "symbol": "MKT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xBEB9eF514a379B997e0798FDcC901Ee474B6D9A1",
            "symbol": "MLN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1a95B271B0535D15fa49932Daba31BA612b52946",
            "symbol": "MNE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xA9877b1e05D035899131DBd1e403825166D09f92",
            "symbol": "MNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x83cee9e086a77e492ee0bb93c2b0437ad6fdeccc",
            "symbol": "MNTP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x865ec58b06bf6305b886793aa20a2da31d034e68",
            "symbol": "MOC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x957c30aB0426e0C93CD8241E2c60392d08c6aC8e",
            "symbol": "MOD",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x501262281b2ba043e2fbf14904980689cddb0c78",
            "symbol": "MORE",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0x263c618480dbe35c300d8d5ecda19bbb986acaed",
            "symbol": "MOT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfbd0d1c77b501796a35d86cf91d65d9778eee695",
            "symbol": "MOVED",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x44bf22949f9cc84b61b9328a9d885d1b5c806b41",
            "symbol": "MOZO",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0xf453b5b9d4e0b5c62ffb256bb2378cc2bc8e8a89",
            "symbol": "MRK",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x82125AFe01819Dff1535D0D6276d57045291B6c0",
            "symbol": "MRL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x21f0F0fD3141Ee9E11B3d7f13a1028CD515f459c",
            "symbol": "MRP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xAB6CF87a50F17d7F5E1FEaf81B6fE9FfBe8EBF84",
            "symbol": "MRV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x68AA3F232dA9bdC2343465545794ef3eEa5209BD",
            "symbol": "MSP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x905E337c6c8645263D3521205Aa37bf4d034e745",
            "symbol": "MTC (Medical Token Currency)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdfdc0d82d96f8fd40ca0cfb4a288955becec2088",
            "symbol": "MTC (Mesh Network)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xaF4DcE16Da2877f8c9e00544c93B62Ac40631F16",
            "symbol": "MTH",
            "decimal": 5,
            "type": "default"
        },
        {
            "address": "0xF433089366899D83a9f26A773D59ec7eCF30355e",
            "symbol": "MTL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x41dbecc1cdc5517c6f76f6a6e836adbee2754de3",
            "symbol": "MTN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7FC408011165760eE31bE2BF20dAf450356692Af",
            "symbol": "MTR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1e49fF77c355A3e38D6651ce8404AF0E48c5395f",
            "symbol": "MTRc",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0AF44e2784637218dD1D32A322D44e603A8f0c6A",
            "symbol": "MTX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x515669d308f887fd83a471c7764f5d084886d34d",
            "symbol": "MUXE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa849eaae994fb86afa73382e9bd88c2b6b18dc71",
            "symbol": "MVL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8a77e40936bbc27e80e9a3f526368c967869c86d",
            "symbol": "MVP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6425c6be902d692ae2db752b3c268afadb099d3b",
            "symbol": "MWAT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf7e983781609012307f2514f63D526D83D24F466",
            "symbol": "MYD",
            "decimal": 16,
            "type": "default"
        },
        {
            "address": "0xa645264C5603E96c3b0B078cdab68733794B0A71",
            "symbol": "MYST",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x8d80de8A78198396329dfA769aD54d24bF90E7aa",
            "symbol": "NAC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xffe02ee4c69edf1b340fcad64fbd6b37a7b9e265",
            "symbol": "NANJ",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x5d65D971895Edc438f465c17DB6992698a52318D",
            "symbol": "NAS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x588047365df5ba589f923604aac23d673555c623",
            "symbol": "NAVI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x17f8aFB63DfcDcC90ebE6e84F060Cc306A98257D",
            "symbol": "NBAI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9f195617fa8fbad9540c5d113a99a0a0172aaedc",
            "symbol": "NBC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x809826cceab68c387726af962713b64cb5cb3cca",
            "symbol": "NCASH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5d48f293baed247a2d0189058ba37aa238bd4725",
            "symbol": "NCC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9e46a38f5daabe8683e10793b06749eef7d733d1",
            "symbol": "NCT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa54ddc7b3cce7fc8b1e3fa0256d0db80d2c10970",
            "symbol": "NDC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcc80c051057b774cd75067dc48f8987c4eb97a5e",
            "symbol": "NEC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd8446236FA95b9b5f9fd0f8E7Df1a944823c683d",
            "symbol": "NEEO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcfb98637bcae43C13323EAa1731cED2B716962fD",
            "symbol": "NET",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa823e6722006afe99e91c30ff5295052fe6b8e32",
            "symbol": "NEU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x814964b1bceAf24e26296D031EaDf134a2Ca4105",
            "symbol": "NEWB",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xb62132e35a6c13ee1ee0f84dc5d40bad8d815206",
            "symbol": "NEXO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x72dd4b6bd852a3aa172be4d6c5a6dbec588cf131",
            "symbol": "NGC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe26517A9967299453d3F1B48Aa005E6127e67210",
            "symbol": "NIMFA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5554e04e76533e1d14c52f05beef6c9d329e1e30",
            "symbol": "NIO",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x1776e1F26f98b1A5dF9cD347953a26dd3Cb46671",
            "symbol": "NMR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x58a4884182d9e835597f405e5f258290e46ae7c2",
            "symbol": "NOAH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf4faea455575354d2699bc209b0a65ca99f69982",
            "symbol": "NOBS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xec46f8207d766012454c408de210bcbc2243e71c",
            "symbol": "NOX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4ce6b362bc77a24966dda9078f9cef81b3b886a7",
            "symbol": "NPER",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x28b5e12cce51f15594b0b91d5b5adaa70f684a02",
            "symbol": "NPX",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0xa15c7ebe1f07caf6bff097d8a589fb8ac49ae5b3",
            "symbol": "NPXS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x69beab403438253f13b6e92db91f7fb849258263",
            "symbol": "NTK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5d4d57cd06fa7fe99e26fdc481b468f77f05073c",
            "symbol": "NTK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8a99ed8a1b204903ee46e733f2c1286f6d20b177",
            "symbol": "NTO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2233799ee2683d75dfefacbcd2a26c78d34b470d",
            "symbol": "NTWK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x245ef47d4d0505ecf3ac463f4d81f41ade8f1fd1",
            "symbol": "NUG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb91318f35bdb262e9423bc7c7c2a3a93dd93c92c",
            "symbol": "NULS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x57Ab1E02fEE23774580C119740129eAC7081e9D3",
            "symbol": "nUSD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7627de4b93263a6a7570b8dafa64bae812e5c394",
            "symbol": "NXX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x5c6183d10A00CD747a6Dbb5F658aD514383e9419",
            "symbol": "NXX OLD",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x45e42D659D9f9466cD5DF622506033145a9b89Bc",
            "symbol": "NxC",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x5e888B83B7287EED4fB7DA7b7d0A0D4c735d94b3",
            "symbol": "OAK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x701C244b988a513c945973dEFA05de933b23Fe1D",
            "symbol": "OAX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0235fe624e044a05eed7a43e16e3083bc8a4287a",
            "symbol": "OCC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4092678e4e78230f46a1534c0fbc8fa39780892b",
            "symbol": "OCN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbf52f2ab39e26e0951d2a02b49b7702abe30406a",
            "symbol": "ODE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6f539a9456a5bcb6334a1a41207c3788f5825207",
            "symbol": "OHNI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc66ea802717bfb9833400264dd12c2bceaa34a6d",
            "symbol": "OLD_MKR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9d9223436ddd466fc247e9dbbd20207e640fef58",
            "symbol": "OLE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x64A60493D888728Cf42616e034a0dfEAe38EFCF0",
            "symbol": "OLT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd26114cd6EE289AccF82350c8d8487fedB8A0C07",
            "symbol": "OMG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb5dbc6d3cf380079df3b27135664b6bcf45d1869",
            "symbol": "OMX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xb23be73573bc7e03db6e5dfc62405368716d28a8",
            "symbol": "ONEK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd341d1680eeee3255b8c4c75bcce7eb57f144dae",
            "symbol": "onG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6863be0e7cf7ce860a574760e9020d519a8bdc47",
            "symbol": "ONL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x69c4BB240cF05D51eeab6985Bab35527d04a8C64",
            "symbol": "OPEN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x4355fC160f74328f9b383dF2EC589bB3dFd82Ba0",
            "symbol": "OPT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x832904863978b94802123106e6eb491bdf0df928",
            "symbol": "OPTI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xff56Cc6b1E6dEd347aA0B7676C85AB0B3D08B0FA",
            "symbol": "ORBS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6F59e0461Ae5E2799F1fB3847f05a63B16d0DbF8",
            "symbol": "ORCA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd2fa8f92ea72abb35dbd6deca57173d22db2ba49",
            "symbol": "ORI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x516E5436bAfdc11083654DE7Bb9b95382d08d5DE",
            "symbol": "ORME",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xeb9a4b185816c354db92db09cc3b50be60b901b6",
            "symbol": "ORS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2C4e8f2D746113d0696cE89B35F0d8bF88E0AEcA",
            "symbol": "OST",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x881ef48211982d01e2cb7092c915e647cd40d85c",
            "symbol": "OTN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x170b275ced089fffaebfe927f445a350ed9160dc",
            "symbol": "OWN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x65a15014964f2102ff58647e16a16a6b9e14bcf6",
            "symbol": "Ox Fina",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0xb9bb08ab7e9fa0a1356bd4a39ec0ca267e03b0b3",
            "symbol": "PAI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfeDAE5642668f8636A11987Ff386bfd215F942EE",
            "symbol": "PAL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xea5f88e54d982cbb0c441cde4e79bc305e5b43bc",
            "symbol": "PARETO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6dd4e4aad29a40edd6a409b9c1625186c9855b4d",
            "symbol": "PARKGENE) (GENE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x77761e63c05aee6648fdaeaa9b94248351af9bcd",
            "symbol": "PASS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xF3b3Cad094B89392fcE5faFD40bC03b80F2Bc624",
            "symbol": "PAT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x694404595e3075a942397f466aacd462ff1a7bd0",
            "symbol": "PATENTS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xF813F3902bBc00A6DCe378634d3B79D84F9803d7",
            "symbol": "PATH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8e870d67f660d95d5be530380d0ec0bd388289e1",
            "symbol": "PAX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB97048628DB6B661D4C2aA833e95Dbe1A905B280",
            "symbol": "PAY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x55648de19836338549130b1af587f16bea46f66b",
            "symbol": "PBL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xF4c07b1865bC326A3c01339492Ca7538FD038Cc0",
            "symbol": "PBT",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0xfcAC7A7515e9A9d7619fA77A1fa738111f66727e",
            "symbol": "PCH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe3f4b4a5d91e5cb9435b947f090a319737036312",
            "symbol": "PCH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3618516F45CD3c913F81F9987AF41077932Bc40d",
            "symbol": "PCL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x53148Bb4551707edF51a1e8d7A93698d18931225",
            "symbol": "PCLOLD",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x0db03B6CDe0B2d427C64a04FeAfd825938368f1F",
            "symbol": "PDATA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8Ae56a6850a7cbeaC3c3Ab2cB311e7620167eAC8",
            "symbol": "PEG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5884969Ec0480556E11d119980136a4C17eDDEd1",
            "symbol": "PET",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xec18f898b4076a3e18f1089d33376cc380bde61d",
            "symbol": "PETRO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x55c2A0C171D920843560594dE3d6EEcC09eFc098",
            "symbol": "PEXT",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x2fa32a39fc1c399e0cc7b2935868f5165de7ce97",
            "symbol": "PFR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x13c2fab6354d3790d8ece4f0f1a3280b4a25ad96",
            "symbol": "PHI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE64509F0bf07ce2d29A7eF19A8A9bc065477C1B4",
            "symbol": "PIPL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x8eFFd494eB698cc399AF6231fCcd39E08fd20B15",
            "symbol": "PIX",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x02f2d4a04e6e01ace88bd2cd632875543b2ef577",
            "symbol": "PKG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2604fa406be957e542beb89e6754fcde6815e83f",
            "symbol": "PKT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x59416A25628A76b4730eC51486114c32E0B582A1",
            "symbol": "PLASMA",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xE477292f1B3268687A29376116B0ED27A9c76170",
            "symbol": "PLAY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0AfFa06e7Fbe5bC9a764C979aA66E8256A631f02",
            "symbol": "PLBT",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xe3818504c1B32bF1557b16C238B2E01Fd3149C17",
            "symbol": "PLR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xD8912C10681D8B21Fd3742244f44658dBA12264E",
            "symbol": "PLU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x846c66cf71c43f80403b51fe3906b3599d63336f",
            "symbol": "PMA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x81b4d08645da11374a03749ab170836e4e539767",
            "symbol": "PMNT",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x93ed3fbe21207ec2e8f2d3c3de6e058cb73bc04d",
            "symbol": "PNK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6758b7d441a9739b98552b373703d8d3d14f9e62",
            "symbol": "POA20",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0e0989b1f9b8a38983c2ba8053269ca62ec9b195",
            "symbol": "POE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x43f6a1be992dee408721748490772b15143ce0a7",
            "symbol": "POIN",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x705EE96c1c160842C92c1aeCfCFfccc9C412e3D9",
            "symbol": "POLL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9992eC3cF6A55b00978cdDF2b27BC6882d88D1eC",
            "symbol": "POLY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x779B7b713C86e3E6774f5040D9cCC2D43ad375F8",
            "symbol": "POOL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xee609fe292128cad03b786dbb9bc2634ccdbe7fc",
            "symbol": "POS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x595832f8fc6bf59c85c527fec3740a1b7a361269",
            "symbol": "POWR",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xc42209accc14029c1012fb5680d95fbd6036e2a0",
            "symbol": "PPP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd4fa1460F537bb9085d22C7bcCB5DD450Ef28e3a",
            "symbol": "PPT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x88a3e4f35d64aad41a6d4030ac9afe4356cb84fa",
            "symbol": "PRE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7728dfef5abd468669eb7f9b48a7f70a501ed29d",
            "symbol": "PRG",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x3adfc4999f77d04c8341bac5f3a76f58dff5b37a",
            "symbol": "PRIX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1844b21593262668b7248d0f57a220caaba46ab9",
            "symbol": "PRL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9041fe5b3fdea0f5e4afdc17e75180738d877a01",
            "symbol": "PRO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x226bb599a12C826476e3A771454697EA52E9E220",
            "symbol": "PRO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xA3149E0fA0061A9007fAf307074cdCd290f0e2Fd",
            "symbol": "PRON",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xE40C374d8805b1dD58CDcEFf998A2F6920Cb52FD",
            "symbol": "PRPS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x163733bcc28dbf26B41a8CfA83e369b5B3af741b",
            "symbol": "PRS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0c04d4f331da8df75f9e2e271e3f3f1494c66c36",
            "symbol": "PRSP",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x5d4abc77b8405ad177d8ac6682d584ecbfd46cec",
            "symbol": "PST",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x66497a283e0a007ba3974e837784c6ae323447de",
            "symbol": "PT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2a8E98e256f32259b5E5Cb55Dd63C8e891950666",
            "symbol": "PTC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8Ae4BF2C33a8e667de34B54938B0ccD03Eb8CC06",
            "symbol": "PTOY",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x4689a4e169eb39cc9078c0940e21ff1aa8a39b9c",
            "symbol": "PTT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5512e1d6a7be424b4323126b4f9e86d023f95764",
            "symbol": "PTWO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xef6b4ce8c9bc83744fbcde2657b32ec18790458a",
            "symbol": "PUC",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xc14830e53aa344e8c14603a91229a0b925b0b262",
            "symbol": "PXT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x7703c35cffdc5cda8d27aa3df2f9ba6964544b6e",
            "symbol": "PYLNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x618e75ac90b12c6049ba3b27f5d5f8651b0037f6",
            "symbol": "QASH",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x671AbBe5CE652491985342e85428EB1b07bC6c64",
            "symbol": "QAU",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xcb5ea3c190d8f82deadf7ce5af855ddbf33e3962",
            "symbol": "QBIT",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x2467aa6b5a2351416fd4c3def8462d841feeecec",
            "symbol": "QBX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xea26c4ac16d4a5a106820bc8aee85fd0b7b2b664",
            "symbol": "QKC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4a220E6096B25EADb88358cb44068A3248254675",
            "symbol": "QNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xFFAA5ffc455d9131f8A2713A741fD1960330508B",
            "symbol": "QRG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x697beac28B09E122C4332D163985e8a73121b97F",
            "symbol": "QRL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x99ea4dB9EE77ACD40B119BD1dC4E33e1C070b80d",
            "symbol": "QSP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2C3C1F05187dBa7A5f2Dd47Dca57281C4d4F183F",
            "symbol": "QTQ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9a642d6b3368ddc662CA244bAdf32cDA716005BC",
            "symbol": "QTUM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x264dc2dedcdcbb897561a57cba5085ca416fb7b4",
            "symbol": "QUN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1183f92a5624d68e85ffb9170f16bf0443b4c242",
            "symbol": "QVT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x48f775efbe4f5ece6e0df2f7b5932df56823b990",
            "symbol": "R",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x45edb535942a8c84d9f4b5d37e1b25f91ea4804c",
            "symbol": "RAO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfc2c4d8f95002c14ed0a7aa65102cac9e5953b5e",
            "symbol": "RBLX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf970b8e36e23f7fc3fd752eea86f8be8d83375a6",
            "symbol": "RCN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x13f25cd52b21650caa8225c9942337d914c9b030",
            "symbol": "RCT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x255aa6df07540cb5d3d297f0d0d4d84cb52bc8e6",
            "symbol": "RDN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x767bA2915EC344015a7938E3eEDfeC2785195D05",
            "symbol": "REA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9214ec02cb71cba0ada6896b8da260736a67ab10",
            "symbol": "REAL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5f53f7a8075614b699baad0bc2c899f4bad8fbbf",
            "symbol": "REBL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x76960dccd5a1fe799f7c29be9f19ceb4627aeb2f",
            "symbol": "RED",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB563300A3BAc79FC09B93b6F84CE0d4465A2AC27",
            "symbol": "REDC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x89303500a7abfb178b274fd89f2469c264951e1f",
            "symbol": "REF",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x83984d6142934bb535793a82adb0a46ef0f66b6d",
            "symbol": "REM",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x408e41876cCCDC0F92210600ef50372656052a38",
            "symbol": "REN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1985365e9f78359a9B6AD760e32412f4a445E862",
            "symbol": "REP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8f8221aFbB33998d8584A2B05749bA73c37a938a",
            "symbol": "REQ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf05a9382A4C3F29E2784502754293D88b835109C",
            "symbol": "REX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd0929d411954c47438dc1d871dd6081f5c5e149c",
            "symbol": "RFR",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x4c383bdcae52a6e1cb810c76c70d6f31a249ec9b",
            "symbol": "RGS",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x168296bb09e24a88805cb9c33356536b980d3fc5",
            "symbol": "RHOC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x9469D013805bFfB7D3DEBe5E7839237e535ec483",
            "symbol": "RING",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdd007278b667f6bef52fd0a4c23604aa1f96039a",
            "symbol": "RIPT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x0b1724cc9fda0186911ef6a75949e9c0d3f0f2f3",
            "symbol": "RIYA",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x106aa49295b525fcf959aa75ec3f7dcbf5352f1c",
            "symbol": "RKT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x607F4C5BB672230e8672085532f7e901544a7375",
            "symbol": "RLC",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0xcCeD5B8288086BE8c38E23567e684C3740be4D48",
            "symbol": "RLT",
            "decimal": 10,
            "type": "default"
        },
        {
            "address": "0xbe99B09709fc753b09BCf557A992F6605D5997B0",
            "symbol": "RLTY",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x4a42d2c580f83dce404acad18dab26db11a1750e",
            "symbol": "RLX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7dc4f41294697a7903c4027f6ac528c5d14cd7eb",
            "symbol": "RMC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x8d5682941ce456900b12d47ac06a88b47c764ce1",
            "symbol": "RMESH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0996bfb5d057faa237640e2506be7b4f9c46de0b",
            "symbol": "RNDR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xff603f43946a3a28df5e6a73172555d8c8b02386",
            "symbol": "RNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1fe70be734e473e5721ea57c8b5b01e6caa52686",
            "symbol": "RNTB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1bcbc54166f6ba149934870b60506199b6c9db6d",
            "symbol": "ROC",
            "decimal": 10,
            "type": "default"
        },
        {
            "address": "0xA40106134c5bF4c41411554e6db99B95A15ed9d8",
            "symbol": "ROCK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xC16b542ff490e01fcc0DC58a60e1EFdc3e357cA6",
            "symbol": "ROCK2",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x0E3de3B0E3D617FD8D1D8088639bA877feb4d742",
            "symbol": "ROCK2PAY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc9de4b7f0c3d991e967158e4d4bfa4b51ec0b114",
            "symbol": "ROK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4993CB95c7443bdC06155c5f5688Be9D8f6999a5",
            "symbol": "ROUND",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb4efd85c19999d84251304bda99e90b92300bd93",
            "symbol": "RPL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xec491c1088eae992b7a214efb0a266ad0927a72a",
            "symbol": "RTB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3fd8f39a962efda04956981c31ab89fab5fb8bc8",
            "symbol": "RTH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x54b293226000ccBFC04DF902eEC567CB4C35a903",
            "symbol": "RTN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf278c1ca969095ffddded020290cf8b5c424ace2",
            "symbol": "RUFF",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdEE02D94be4929d26f67B64Ada7aCf1914007F10",
            "symbol": "RUNE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3d1ba9be9f66b8ee101911bc36d3fb562eac2244",
            "symbol": "RVT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe8663a64a96169ff4d95b4299e7ae9a76b905b31",
            "symbol": "Rating",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x1ec8fe51a9b6a3a6c427d17d9ecc3060fbc4a45c",
            "symbol": "S-A-PAT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3eb91d237e491e0dee8582c402d85cb440fb6b54",
            "symbol": "S-ETH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4156D3342D5c385a87D264F90653733592000581",
            "symbol": "SALT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x7C5A0CE9267ED19B22F8cae653F198e3E8daf098",
            "symbol": "SAN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x78fe18e41f436e1981a3a60d1557c8a7a9370461",
            "symbol": "SCANDI",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0xd7631787b4dcc87b1254cfd1e5ce48e96823dee8",
            "symbol": "SCL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x24dcc881e7dd730546834452f21872d5cb4b5293",
            "symbol": "SCRL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa13f0743951b4f6e3e3aa039f682e17279f52bc3",
            "symbol": "SENC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6745fAB6801e376cD24F03572B9C9B0D4EdDDCcf",
            "symbol": "SENSE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xa44e5137293e855b1b7bc7e2c6f8cd796ffcb037",
            "symbol": "SENT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xe06eda7435ba749b047380ced49121dde93334ae",
            "symbol": "SET",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x98f5e9b7f0e33956c0443e81bf7deb8b5b1ed545",
            "symbol": "SEXY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa1ccc166faf0E998b3E33225A1A0301B1C86119D",
            "symbol": "SGEL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb2135ab9695a7678dd590b1a996cb0f37bcb0718",
            "symbol": "SGN",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x33c623a2baafeb8d15dfaf3ce44095efec83d72c",
            "symbol": "SGP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcb5a05bef3257613e984c17dbcf039952b6d883f",
            "symbol": "SGR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x37427576324fE1f3625c9102674772d7CF71377d",
            "symbol": "SGT (SelfieYo Gold Token)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd248B0D48E44aaF9c49aea0312be7E13a6dc1468",
            "symbol": "SGT (StatusGenesis)",
            "decimal": 1,
            "type": "default"
        },
        {
            "address": "0xe25b0bba01dc5630312b6a21927e578061a13f55",
            "symbol": "SHIP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xEF2E9966eb61BB494E5375d5Df8d67B7dB8A780D",
            "symbol": "SHIT",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x8542325b72c6d9fc0ad2ca965a78435413a915a0",
            "symbol": "SHL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xef2463099360a085f1f10b076ed72ef625497a06",
            "symbol": "SHP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8a187d5285d316bcbc9adafc08b51d70a0d8e000",
            "symbol": "SIFT",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x6888a16eA9792c15A4DCF2f6C623D055c8eDe792",
            "symbol": "SIG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4af328c52921706dcb739f25786210499169afe6",
            "symbol": "SKB",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x2bDC0D42996017fCe214b21607a515DA41A9E0C5",
            "symbol": "SKIN",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xd99b8a7fa48e25cce83b81812220a3e03bf64e5f",
            "symbol": "SKM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4994e81897a920c0FEA235eb8CEdEEd3c6fFF697",
            "symbol": "SKO1",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4c382F8E09615AC86E08CE58266CC227e7d4D913",
            "symbol": "SKR",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xfdFE8b7aB6CF1bD1E3d14538Ef40686296C42052",
            "symbol": "SKRP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6E34d8d84764D40f6D7b39cd569Fd017bF53177D",
            "symbol": "SKRP (Phase 1)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x324a48ebcbb46e61993931ef9d35f6697cd2901b",
            "symbol": "SKRP (Phase 1-E)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7A5fF295Dc8239d5C2374E4D894202aAF029Cab6",
            "symbol": "SLT",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x7928c8aBF1F74eF9F96D4D0a44e3b4209d360785",
            "symbol": "SLY",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6F6DEb5db0C4994A8283A01D6CFeEB27Fc3bBe9C",
            "symbol": "SMART",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x39013f961c378f02c2b82a6e1d31e9812786fd9d",
            "symbol": "SMS",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x2dcfaac11c9eebd8c6c42103fe9e2a6ad237af27",
            "symbol": "SMT (Smart Node)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x55f93985431fc9304077687a35a1ba103dc1e081",
            "symbol": "SMT (SmartMesh)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x78Eb8DC641077F049f910659b6d580E80dC4d237",
            "symbol": "SMT (Social Media Market)",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x198a87b3114143913d4229fb0f6d4bcb44aa8aff",
            "symbol": "SNBL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xF4134146AF2d511Dd5EA8cDB1C4AC88C57D60404",
            "symbol": "SNC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf333b2Ace992ac2bBD8798bF57Bc65a06184afBa",
            "symbol": "SND",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xcFD6Ae8BF13f42DE14867351eAff7A8A3b9FbBe7",
            "symbol": "SNG",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xaeC2E87E0A235266D9C5ADc9DEb4b2E29b54D009",
            "symbol": "SNGLS",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x44F588aEeB8C44471439D1270B3603c66a9262F1",
            "symbol": "SNIP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x983F6d60db79ea8cA4eB9968C6aFf8cfA04B3c63",
            "symbol": "SNM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbdc5bac39dbe132b1e030e898ae3830017d7d969",
            "symbol": "SNOV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x744d70FDBE2Ba4CF95131626614a1763DF805B9E",
            "symbol": "SNT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2859021ee7f2cb10162e67f33af2d22764b31aff",
            "symbol": "SNTR",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x2d0e95bd4795d7ace0da3c0ff7b706a5970eb9d3",
            "symbol": "SOC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1f54638b7737193ffd86c19ec51907a7c41755d8",
            "symbol": "SOL",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x1c62aca2b7605db3606eacda7bc67a1857ddb8ff",
            "symbol": "SONIQ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x42d6622dece394b54999fbd73d108123806f6a18",
            "symbol": "SPANK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x58bf7df57d9DA7113c4cCb49d8463D4908C735cb",
            "symbol": "SPARC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x24aef3bf1a47561500f9430d74ed4097c47f51f2",
            "symbol": "SPARTA",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x1dea979ae76f26071870f824088da78979eb91c8",
            "symbol": "SPD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x85089389C14Bd9c77FC2b8F0c3d1dC3363Bf06Ef",
            "symbol": "SPF",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x3833dda0aeb6947b98ce454d89366cba8cc55528",
            "symbol": "SPHTX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0324dd195d0cd53f9f07bee6a48ee7a20bad738f",
            "symbol": "SPICE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x20F7A3DdF244dc9299975b4Da1C39F8D5D75f05A",
            "symbol": "SPN",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x05aaaa829afa407d83315cded1d45eb16025910c",
            "symbol": "SPX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x68d57c9a1C35f63E2c83eE8e49A64e9d70528D25",
            "symbol": "SRN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbbFF862d906E348E9946Bfb2132ecB157Da3D4b4",
            "symbol": "SS (Sharder)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6e2050CBFB3eD8A4d39b64cC9f47E711a03a5a89",
            "symbol": "SSH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x624d520bab2e4ad83935fa503fb130614374e850",
            "symbol": "SSP",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x9a005c9a89bd72a4bd27721e7a09a3c11d2b03c4",
            "symbol": "STAC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xF70a642bD387F94380fFb90451C2c81d4Eb82CBc",
            "symbol": "STAR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x09bca6ebab05ee2ae945be4eda51393d94bf7b99",
            "symbol": "STB",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x629aEe55ed49581C33ab27f9403F7992A289ffd5",
            "symbol": "STC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xaE73B38d1c9A8b274127ec30160a4927C4d71824",
            "symbol": "STK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x599346779e90fc3F5F997b5ea715349820F91571",
            "symbol": "STN",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0xB64ef51C888972c908CFacf59B47C1AfBC0Ab8aC",
            "symbol": "STORJ",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xD0a4b8946Cb52f0661273bfbC6fD0E0C75Fc6433",
            "symbol": "STORM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xecd570bBf74761b960Fa04Cc10fe2c4e86FfDA36",
            "symbol": "STP",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x5c3a228510d246b78a3765c20221cbf3082b44a4",
            "symbol": "STQ",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xBAE235823D7255D9D48635cEd4735227244Cd583",
            "symbol": "STR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x46492473755e8dF960F8034877F61732D718CE96",
            "symbol": "STRC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x0371a82e4a9d0a4312f3ee2ac9c6958512891372",
            "symbol": "STU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x006BeA43Baa3f7A6f765F14f10A1a1b08334EF45",
            "symbol": "STX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x12480E24eb5bec1a9D4369CaB6a80caD3c0A377A",
            "symbol": "SUB",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0xe120c1ecbfdfea7f0a8f0ee30063491e8c26fedf",
            "symbol": "SUR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xbdeb4b83251fb146687fa19d1c660f99411eefe3",
            "symbol": "SVD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0bb217e40f8a5cb79adf04e1aab60e5abd0dfc1e",
            "symbol": "SWFTC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x9e88613418cf03dca54d6a2cf6ad934a78c7a17a",
            "symbol": "SWM",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB9e7F8568e08d5659f5D29C4997173d84CdF2607",
            "symbol": "SWT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x12b306fa98f4cbb8d4457fdff3a0a0a56f07ccdf",
            "symbol": "SXDT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2c82c73d5b34aa015989462b2948cd616a37641f",
            "symbol": "SXUT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x10b123fddde003243199aad03522065dc05827a0",
            "symbol": "SYN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb1eef147028e9f480dbc5ccaa3277d417d1b85f0",
            "symbol": "Seele",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4ca74185532dc1789527194e5b9c866dd33f4e82",
            "symbol": "SenSatorI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xD65960FAcb8E4a2dFcb2C2212cb2e44a02e2a57E",
            "symbol": "Soar",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x1d4ccc31dab6ea20f461d329a0562c1c58412515",
            "symbol": "TALAO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc27a2f05fa577a83ba0fdb4c38443c0718356501",
            "symbol": "TAU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xFACCD5Fc83c3E4C3c1AC1EF35D15adf06bCF209C",
            "symbol": "TBC2",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xAFe60511341a37488de25Bef351952562E31fCc1",
            "symbol": "TBT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x3a92bd396aef82af98ebc0aa9030d25a23b11c6b",
            "symbol": "TBX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfA0eF5E034CaE1AE752d59bdb8aDcDe37Ed7aB97",
            "symbol": "TCA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9972a0f24194447e73a7e8b6cd26a52e02ddfad5",
            "symbol": "TCH",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x2a1dbabe65c595B0022e75208C34014139d5d357",
            "symbol": "TDH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1c79ab32c66acaa1e9e81952b8aaa581b43e54e7",
            "symbol": "TEAM",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x85e076361cc813a908ff672f9bad1541474402b2",
            "symbol": "TEL",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0xdd16ec0f66e54d453e6756713e533355989040e4",
            "symbol": "TEN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe5f166c0d8872b68790061317bb6cca04582c912",
            "symbol": "TFD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa7f976C360ebBeD4465c2855684D1AAE5271eFa9",
            "symbol": "TFL",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf8e06e4e4a80287fdca5b02dccecaa9d0954840f",
            "symbol": "TGAME",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xac3da587eac229c9896d919abc235ca4fd7f72c1",
            "symbol": "TGT",
            "decimal": 1,
            "type": "default"
        },
        {
            "address": "0x3883f5e181fccaf8410fa61e12b59bad963fb645",
            "symbol": "THETA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1cb3209d45b2a60b7fbca1ccdbf87f674237a4aa",
            "symbol": "THR",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x4f27053f32eda8af84956437bc00e5ffa7003287",
            "symbol": "THRT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfe7B915A0bAA0E79f85c5553266513F7C1c03Ed0",
            "symbol": "THUG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x72430a612adc007c50e3b6946dbb1bb0fd3101d1",
            "symbol": "TIC",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x7F4B2A690605A7cbb66F7AA6885EbD906a5e2E9E",
            "symbol": "TICO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x999967e2ec8a74b7c8e9db19e039d920b31d39d0",
            "symbol": "TIE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xeee2d00eb7deb8dd6924187f5aa3496b7d06e62a",
            "symbol": "TIG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6531f133e6DeeBe7F2dcE5A0441aA7ef330B4e53",
            "symbol": "TIME",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x80bc5512561c7f85a3a9508c7df7901b370fa1df",
            "symbol": "TIO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xEa1f346faF023F974Eb5adaf088BbCdf02d761F4",
            "symbol": "TIX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xdae1baf249964bc4b6ac98c3122f0e3e785fd279",
            "symbol": "TKA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xaAAf91D9b90dF800Df4F55c205fd6989c977E73a",
            "symbol": "TKN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xb45a50545beeab73f38f31e5973768c421805e5e",
            "symbol": "TKR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb3616550abc8af79c7a5902def9efa3bc9a95200",
            "symbol": "TLX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x3209f98bebf0149b769ce26d71f7aea8e435efea",
            "symbol": "TMT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb0280743b44bf7db4b6be482b2ba7b75e5da096c",
            "symbol": "TNS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x08f5a9235b08173b7569f83645d2c7fb55e8ccd8",
            "symbol": "TNT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x8b353021189375591723e7384262f45709a3c3dc",
            "symbol": "TOMO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8eb965ee9cCFBCE76c0a06264492c0afEfc2826d",
            "symbol": "TOOR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xaa7a9ca87d3694b5755f213b5d04094b8d0f0a6f",
            "symbol": "TRAC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x12759512d326303b45f1cec8f7b6fd96f387778e",
            "symbol": "TRAK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcB3F902bf97626391bF8bA87264bbC3DC13469be",
            "symbol": "TRC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x566Fd7999B1Fc3988022bD38507A48F0bCf22c77",
            "symbol": "TRCN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x30ceCB5461A449A90081F5a5F55db4e048397BAB",
            "symbol": "TRCT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x33f90dee07c6e8b9682dd20f73e6c358b2ed0f03",
            "symbol": "TRDT",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xcb94be6f13a1182e4a4b6140cb7bf2025d28e41b",
            "symbol": "TRST",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xf230b790e05390fc8295f4d3f60332c93bed42e2",
            "symbol": "TRX",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x6B87999bE87358065bBdE41e8a0fe0B7b1cd2514",
            "symbol": "TSW",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xaab606817809841e8b1168be8779eeaf6744ef64",
            "symbol": "TTA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9389434852b94bbad4c8afed5b7bdbc5ff0c2275",
            "symbol": "TTC (TTC)",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9cda8a60dd5afa156c95bd974428d91a0812e054",
            "symbol": "TTU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8dd5fbce2f6a956c3022ba3663759011dd51e73e",
            "symbol": "TUSD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2eF1aB8a26187C58BB8aAeB11B2fC6D25C5c0716",
            "symbol": "TWN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE7775A6e9Bcf904eb39DA2b68c5efb4F9360e08C",
            "symbol": "TaaS",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x8400d94a5cb0fa0d041a3788e395285d61c9ee5e",
            "symbol": "UBT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x92e52a1a235d9a103d970901066ce910aacefd37",
            "symbol": "UCASH",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xaaf37055188feee4869de63464937e683d61b2a1",
            "symbol": "UCN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xea097a2b1db00627b2fa17460ad260c016016977",
            "symbol": "UFR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x24692791bc444c5cd0b81e3cbcaba4b04acd1f3b",
            "symbol": "UKG",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x8e5afc69f6227a3ad75ed346c8723bc62ce97123",
            "symbol": "UMKA",
            "decimal": 4,
            "type": "default"
        },
        {
            "address": "0x6ba460ab75cd2c56343b3517ffeba60748654d26",
            "symbol": "UP",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xc86d054809623432210c107af2e3f619dcfbf652",
            "symbol": "UPP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd01db73e047855efb414e6202098c4be4cd2423b",
            "symbol": "UQC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x931684139f756C24eC0731E9F74FE50e5548dDeF",
            "symbol": "URB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48",
            "symbol": "USDC",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0xD760ADdFb24D9C01Fe4Bfea7475C5e3636684058",
            "symbol": "USDM",
            "decimal": 2,
            "type": "default"
        },
        {
            "address": "0xdac17f958d2ee523a2206206994597c13d831ec7",
            "symbol": "USDT",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x70a72833d6bf7f508c8224ce59ea1ef3d0ea3a38",
            "symbol": "UTK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9e3319636e2126e3c0bc9e3134AEC5e1508A46c7",
            "symbol": "UTN-P",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x16f812be7fff02caf662b85d5d58a5da6572d4df",
            "symbol": "UTT",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x3543638eD4a9006E4840B105944271Bcea15605D",
            "symbol": "UUU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x89205A3A3b2A69De6Dbf7f01ED13B2108B2c43e7",
            "symbol": "Unicorn",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x57C75ECCc8557136D32619a191fBCDc88560d711",
            "symbol": "VDG",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x82BD526bDB718C6d4DD2291Ed013A5186cAE2DCa",
            "symbol": "VDOC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x340d2bde5eb28c1eed91b2f790723e3b160613b7",
            "symbol": "VEE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xD850942eF8811f2A866692A623011bDE52a462C1",
            "symbol": "VEN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xEbeD4fF9fe34413db8fC8294556BBD1528a4DAca",
            "symbol": "VENUS",
            "decimal": 3,
            "type": "default"
        },
        {
            "address": "0x8f3470A7388c05eE4e7AF3d01D8C722b0FF52374",
            "symbol": "VERI",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x2C974B2d0BA1716E644c1FC59982a89DDD2fF724",
            "symbol": "VIB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe8ff5c9c75deb346acac493c463c8950be03dfba",
            "symbol": "VIBE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x882448f83d90b2bf477af2ea79327fdea1335d93",
            "symbol": "VIBEX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf03f8d65bafa598611c3495124093c56e8f638f0",
            "symbol": "VIEW",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd2946be786f35c3cc402c29b323647abda799071",
            "symbol": "VIKKY",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf3e014fe81267870624132ef3a646b8e83853a96",
            "symbol": "VIN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x23b75Bc7AaF28e2d6628C3f424B3882F8f072a3c",
            "symbol": "VIT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1b793e49237758dbd8b752afc9eb4b329d5da016",
            "symbol": "VITE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x519475b31653e46d20cd09f9fdcf3b12bdacb4f5",
            "symbol": "VIU",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x922ac473a3cc241fd3a0049ed14536452d58d73c",
            "symbol": "VLD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xc3bc9eb71f75ec439a6b6c8e8b746fcf5b62f703",
            "symbol": "VOC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x83eEA00D838f92dEC4D1475697B9f4D3537b56E3",
            "symbol": "VOISE",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xeDBaF3c5100302dCddA53269322f3730b1F0416d",
            "symbol": "VRS",
            "decimal": 5,
            "type": "default"
        },
        {
            "address": "0x92e78dae1315067a8819efd6dca432de9dcde2e9",
            "symbol": "VRS",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x5c543e7AE0A1104f78406C340E9C64FD9fCE5170",
            "symbol": "VSL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x9720b467a710382A232a32F540bDCed7d662a10B",
            "symbol": "VZT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4bbbc57af270138ef2ff2c50dbfad684e9e0e604",
            "symbol": "WAB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x829A4cA1303383F1082B6B1fB937116e4b3b5605",
            "symbol": "WATT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x39Bb259F66E1C59d5ABEF88375979b4D20D98022",
            "symbol": "WAX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x74951B677de32D596EE851A233336926e6A2cd09",
            "symbol": "WBA",
            "decimal": 7,
            "type": "default"
        },
        {
            "address": "0x6a0a97e47d15aad1d132a1ac79a480e3f2079063",
            "symbol": "WCT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x840fe75abfadc0f2d54037829571b2782e919ce4",
            "symbol": "WEB",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
            "symbol": "WETH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xf4fe95603881d0e07954fd7605e0e9a916e42c44",
            "symbol": "WHEN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe933c0Cd9784414d5F278C114904F5A84b396919",
            "symbol": "WHO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x62cd07d414ec50b68c7ecaa863a23d344f2d062f",
            "symbol": "WIC",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xD3C00772B24D997A812249ca637a921e81357701",
            "symbol": "WILD",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x899338b84d25ac505a332adce7402d697d947494",
            "symbol": "WIN",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x667088b212ce3d06a1b553a7221E1fD19000d9aF",
            "symbol": "WINGS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1b22c32cd936cb97c28c5690a0695a82abf688e6",
            "symbol": "WISH",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xF6B55acBBC49f4524Aa48D19281A9A77c54DE10f",
            "symbol": "WLK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xbfbe5332f172d77811bc6c272844f3e54a7b23bb",
            "symbol": "WMK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd73A66B8FB26Be8B0AcD7c52Bd325054Ac7d468b",
            "symbol": "WNK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x728781E75735dc0962Df3a51d7Ef47E798A7107E",
            "symbol": "WOLK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa686514faf7d54289266f483d1e4852c99e13ec7",
            "symbol": "WORK",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x4CF488387F035FF08c371515562CBa712f9015d4",
            "symbol": "WPR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x72adadb447784dd7ab1f472467750fc485e4cb2d",
            "symbol": "WRC",
            "decimal": 6,
            "type": "default"
        },
        {
            "address": "0x71e8d74ff1c923e369d0e70dfb09866629c4dd35",
            "symbol": "WRK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb7cb1c96db6b22b0d3d9536e0108d062bd488f74",
            "symbol": "WTC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x84119cb33e8f590d75c2d6ea4e6b0741a7494eda",
            "symbol": "WTT",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0xd8950fDeaa10304B7A7Fd03a2FC66BC39f3c711a",
            "symbol": "WYS",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x056017c55aE7AE32d12AeF7C679dF83A85ca75Ff",
            "symbol": "WYV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x286BDA1413a2Df81731D4930ce2F862a35A609fE",
            "symbol": "WaBi",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x5e4ABE6419650CA839Ce5BB7Db422b881a6064bB",
            "symbol": "WiC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x910Dfc18D6EA3D6a7124A6F8B5458F281060fa4c",
            "symbol": "X8X",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4DF812F6064def1e5e029f1ca858777CC98D2D81",
            "symbol": "XAUR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x49aec0752e68d0282db544c677f6ba407ba17ed7",
            "symbol": "XBL",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x28dee01d53fed0edf5f6e310bf8ef9311513ae40",
            "symbol": "XBP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x4d829f8c92a6691c56300d020c9e0db984cfe2ba",
            "symbol": "XCC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x1e26b3d07e57f453cae30f7ddd2f945f5bf3ef33",
            "symbol": "XCLR",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x41ab1b6fcbb2fa9dced81acbdec13ea6315f2bf2",
            "symbol": "XDCE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa017ac5fac5941f95010b12570b812c974469c2c",
            "symbol": "XES",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x054c64741dbafdc19784505494029823d89c3b13",
            "symbol": "XET",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x16aF5bfb4Ae7E475b9aDC3Bf5Cb2f1E6a50d7940",
            "symbol": "XFS",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xf6b6aa0ef0f5edc2c1c5d925477f97eaf66303e7",
            "symbol": "XGG",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x533ef0984b2FAA227AcC620C67cce12aA39CD8CD",
            "symbol": "XGM",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x30f4A3e0aB7a76733D8b60b89DD93c3D0b4c9E2f",
            "symbol": "XGT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB110eC7B1dcb8FAB8dEDbf28f53Bc63eA5BEdd84",
            "symbol": "XID",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x44449Fa4d607F807d1eD4a69ad942971728391C8",
            "symbol": "XMCT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0f8c45b896784a1e408526b9300519ef8660209c",
            "symbol": "XMX",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xBC86727E770de68B1060C91f6BB6945c73e10388",
            "symbol": "XNK",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xab95e915c123fded5bdfb6325e35ef5515f1ea69",
            "symbol": "XNN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x572e6f318056ba0c5d47a422653113843d250691",
            "symbol": "XNT",
            "decimal": 0,
            "type": "default"
        },
        {
            "address": "0x153ed9cc1b792979d2bde0bbf45cc2a7e436a5f9",
            "symbol": "XOV",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x90528aeb3a2b736b780fd1b6c478bb7e1d643170",
            "symbol": "XPA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xBB1fA4FdEB3459733bF67EbC6f893003fA976a82",
            "symbol": "XPAT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xB24754bE79281553dc1adC160ddF5Cd9b74361a4",
            "symbol": "XRL",
            "decimal": 9,
            "type": "default"
        },
        {
            "address": "0x0F513fFb4926ff82D7F60A05069047AcA295C413",
            "symbol": "XSC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x55296f69f40ea6d20e478533c15a6b08b654e758",
            "symbol": "XYO",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x922105fad8153f516bcfb829f56dc097a0e1d705",
            "symbol": "YEE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xca2796f9f61dc7b238aab043971e49c6164df375",
            "symbol": "YEED",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xcbeaec699431857fdb4d37addbbdc20e132d4903",
            "symbol": "YOYOW",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xd9a12cde03a86e800496469858de8581d3a5353d",
            "symbol": "YUP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x0F33bb20a282A7649C7B3AFf644F084a9348e933",
            "symbol": "YUPIE",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x6781a0f84c7e9e846dcb84a9a5bd49333067b104",
            "symbol": "ZAP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb9EF770B6A5e12E45983C5D80545258aA38F3B78",
            "symbol": "ZCN",
            "decimal": 10,
            "type": "default"
        },
        {
            "address": "0x2008e3057bd734e10ad13c9eae45ff132abc1722",
            "symbol": "ZCO",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0x05f4a42e251f2d52b8ed15E9FEdAacFcEF1FAD27",
            "symbol": "ZIL",
            "decimal": 12,
            "type": "default"
        },
        {
            "address": "0x4aac461c86abfa71e9d00d9a2cde8d74e4e1aeea",
            "symbol": "ZINC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xa9d2927d3a04309e008b6af6e2e282ae2952e7fd",
            "symbol": "ZIP",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xedd7c94fd7b4971b916d15067bc454b9e1bad980",
            "symbol": "ZIPT",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xfd8971d5e8e1740ce2d0a84095fca4de729d0c16",
            "symbol": "ZLA",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x554ffc77f4251a9fb3c0e3590a6a205f8d4e067d",
            "symbol": "ZMN",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xb5b8f5616fe42d5ceca3e87f3fddbdd8f496d760",
            "symbol": "ZPR",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xE41d2489571d322189246DaFA5ebDe1F4699F498",
            "symbol": "ZRX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x7A41e0517a5ecA4FdbC7FbebA4D4c47B9fF6DC63",
            "symbol": "ZSC",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0xe386b139ed3715ca4b18fd52671bdcea1cdfe4b1",
            "symbol": "ZST",
            "decimal": 8,
            "type": "default"
        },
        {
            "address": "0xE8F9fa977ea585591d9F394681318C16552577fB",
            "symbol": "ZTX",
            "decimal": 18,
            "type": "default"
        },
        {
            "address": "0x83e2be8d114f9661221384b3a50d24b96a5653f5",
            "symbol": "ZXC",
            "decimal": 18,
            "type": "default"
        }
    ]
        
    #print('Length = ',len(known_token_symbols))
    
    for i in range(len(known_ERC20_token_symbols)):
        if known_ERC20_token_symbols[i]['symbol'] == token_symbol:
            return known_ERC20_token_symbols[i]['address']
        
    return None
